"""Classes to extend Qt GUI with widgets."""

from __future__ import annotations

import sys

from prettyqt.qt.QtWidgets import *  # noqa: F403

from .style import Style, StyleMixin
from .commonstyle import CommonStyle, CommonStyleMixin
from .proxystyle import ProxyStyle
from .application import Application
from .sizepolicy import SizePolicy
from .widget import Widget, WidgetMixin
from .frame import Frame, FrameMixin
from .focusframe import FocusFrame
from .abstractslider import AbstractSlider, AbstractSliderMixin
from .abstractscrollarea import AbstractScrollArea, AbstractScrollAreaMixin
from .abstractbutton import AbstractButton, AbstractButtonMixin
from .lineedit import LineEdit
from .abstractspinbox import AbstractSpinBox, AbstractSpinBoxMixin
from .abstractitemview import AbstractItemView, AbstractItemViewMixin
from .scrollbar import ScrollBar
from .scrollarea import ScrollArea
from .rubberband import RubberBand
from .graphicstransform import GraphicsTransform, GraphicsTransformMixin
from .graphicsrotation import GraphicsRotation
from .graphicsscale import GraphicsScale
from .graphicsitem import GraphicsItem, GraphicsItemMixin
from .graphicsitemgroup import GraphicsItemGroup
from .abstractgraphicsshapeitem import (
    AbstractGraphicsShapeItem,
    AbstractGraphicsShapeItemMixin,
)
from .graphicspixmapitem import GraphicsPixmapItem
from .graphicsobject import GraphicsObject, GraphicsObjectMixin
from .graphicstextitem import GraphicsTextItem
from .graphicslayoutitem import GraphicsLayoutItem, GraphicsLayoutItemMixin
from .graphicslayout import GraphicsLayout, GraphicsLayoutMixin
from .graphicsgridlayout import GraphicsGridLayout
from .graphicslinearlayout import GraphicsLinearLayout
from .graphicsanchorlayout import GraphicsAnchorLayout
from .graphicswidget import GraphicsWidget, GraphicsWidgetMixin
from .graphicsproxywidget import GraphicsProxyWidget
from .graphicslineitem import GraphicsLineItem
from .graphicsrectitem import GraphicsRectItem
from .graphicssimpletextitem import GraphicsSimpleTextItem
from .graphicspolygonitem import GraphicsPolygonItem
from .graphicsellipseitem import GraphicsEllipseItem
from .graphicspathitem import GraphicsPathItem
from .graphicseffect import GraphicsEffect, GraphicsEffectMixin
from .graphicsblureffect import GraphicsBlurEffect
from .graphicscolorizeeffect import GraphicsColorizeEffect
from .graphicsdropshadoweffect import GraphicsDropShadowEffect
from .graphicsopacityeffect import GraphicsOpacityEffect
from .graphicsscene import GraphicsScene
from .graphicsview import GraphicsView, GraphicsViewMixin
from .styleoption import StyleOption, StyleOptionMixin
from .styleoptionbutton import StyleOptionButton
from .styleoptioncomplex import StyleOptionComplex, StyleOptionComplexMixin
from .styleoptiondockwidget import StyleOptionDockWidget
from .styleoptionfocusrect import StyleOptionFocusRect
from .styleoptionframe import StyleOptionFrame
from .styleoptiongraphicsitem import StyleOptionGraphicsItem
from .styleoptionheader import StyleOptionHeader
from .styleoptionmenuitem import StyleOptionMenuItem
from .styleoptionprogressbar import StyleOptionProgressBar
from .styleoptionrubberband import StyleOptionRubberBand
from .styleoptiontab import StyleOptionTab
from .styleoptiontabbarbase import StyleOptionTabBarBase
from .styleoptiontabwidgetframe import StyleOptionTabWidgetFrame
from .styleoptiontoolbar import StyleOptionToolBar
from .styleoptiontoolbox import StyleOptionToolBox
from .styleoptionviewitem import StyleOptionViewItem

from .styleoptioncombobox import StyleOptionComboBox
from .styleoptiongroupbox import StyleOptionGroupBox
from .styleoptionsizegrip import StyleOptionSizeGrip
from .styleoptionslider import StyleOptionSlider
from .styleoptionspinbox import StyleOptionSpinBox
from .styleoptiontitlebar import StyleOptionTitleBar
from .styleoptiontoolbutton import StyleOptionToolButton

from .stylepainter import StylePainter
from .stylefactory import StyleFactory
from .pushbutton import PushButton, PushButtonMixin
from .dialogbuttonbox import DialogButtonBox
from .dialog import Dialog, DialogMixin
from .messagebox import MessageBox
from .errormessage import ErrorMessage

from .fileiconprovider import FileIconProvider
from .filesystemmodel import FileSystemModel

from .slider import Slider
from .dial import Dial

from .dockwidget import DockWidget

from .widgetaction import WidgetAction
from .menu import Menu
from .mainwindow import MainWindow
from .whatsthis import WhatsThis
from .listwidgetitem import ListWidgetItem
from .treewidgetitem import TreeWidgetItem
from .toolbutton import ToolButton
from .tooltip import ToolTip
from .menubar import MenuBar
from .statusbar import StatusBar
from .tabbar import TabBar
from .tabwidget import TabWidget
from .mdisubwindow import MdiSubWindow
from .mdiarea import MdiArea
from .label import Label
from .toolbar import ToolBar
from .headerview import HeaderView
from .commandlinkbutton import CommandLinkButton
from .radiobutton import RadioButton
from .combobox import ComboBox, ComboBoxMixin
from .fontcombobox import FontComboBox
from .spinbox import SpinBox
from .doublespinbox import DoubleSpinBox
from .checkbox import CheckBox
from .keysequenceedit import KeySequenceEdit
from .textedit import TextEdit, TextEditMixin
from .datetimeedit import DateTimeEdit, DateTimeEditMixin
from .dateedit import DateEdit
from .timeedit import TimeEdit
from .calendarwidget import CalendarWidget
from .plaintextedit import PlainTextEdit, PlainTextEditMixin
from .textbrowser import TextBrowser
from .completer import Completer
from .progressbar import ProgressBar
from .lcdnumber import LCDNumber
from .columnview import ColumnView
from .listview import ListView, ListViewMixin
from .listwidget import ListWidget
from .treeview import TreeView, TreeViewMixin
from .tablewidgetselectionrange import TableWidgetSelectionRange
from .treewidget import TreeWidget
from .treewidgetitemiterator import TreeWidgetItemIterator
from .tableview import TableView, TableViewMixin
from .tablewidgetitem import TableWidgetItem
from .tablewidget import TableWidget
from .scrollerproperties import ScrollerProperties
from .scroller import Scroller
from .abstractitemdelegate import AbstractItemDelegate, AbstractItemDelegateMixin
from .itemdelegate import ItemDelegate
from .styleditemdelegate import StyledItemDelegate
from .systemtrayicon import SystemTrayIcon

from .layoutitem import LayoutItem, LayoutItemMixin
from .widgetitem import WidgetItem
from .layout import Layout, LayoutMixin
from .spaceritem import SpacerItem
from .formlayout import FormLayout
from .boxlayout import BoxLayout
from .hboxlayout import HBoxLayout
from .vboxlayout import VBoxLayout
from .stackedlayout import StackedLayout
from .gridlayout import GridLayout
from .toolbox import ToolBox

from .splashscreen import SplashScreen
from .progressdialog import ProgressDialog
from .fontdialog import FontDialog
from .filedialog import FileDialog
from .colordialog import ColorDialog
from .inputdialog import InputDialog
from .buttongroup import ButtonGroup
from .groupbox import GroupBox
from .splitterhandle import SplitterHandle
from .splitter import Splitter
from .wizard import Wizard
from .wizardpage import WizardPage
from .stackedwidget import StackedWidget

from .undoview import UndoView

from .datawidgetmapper import DataWidgetMapper
from .sizegrip import SizeGrip

from .plaintextdocumentlayout import PlainTextDocumentLayout

from .gesture import Gesture, GestureMixin
from .tapgesture import TapGesture
from .tapandholdgesture import TapAndHoldGesture
from .pangesture import PanGesture
from .pinchgesture import PinchGesture
from .swipegesture import SwipeGesture

from .itemeditorcreatorbase import ItemEditorCreatorBase
from .itemeditorfactory import ItemEditorFactory


# to register delegates
from prettyqt import validators
from prettyqt import itemdelegates
from prettyqt import custom_widgets
from prettyqt import itemmodels
from prettyqt.qt import QtWidgets

QT_MODULE = QtWidgets


def app(args: list[str] | None = None, style: str = "Fusion", **kwargs) -> Application:
    if (instance := Application.instance()) is not None:
        return instance
    #  + ["--ignore-gpu-blacklist", "--enable-gpu-rasterization"]
    app = Application(sys.argv if args is None else args, **kwargs)
    app.set_style(style)
    return app


__all__ = [
    "AbstractButton",
    "AbstractButtonMixin",
    "AbstractGraphicsShapeItem",
    "AbstractGraphicsShapeItemMixin",
    "AbstractItemDelegate",
    "AbstractItemDelegateMixin",
    "AbstractItemView",
    "AbstractItemViewMixin",
    "AbstractScrollArea",
    "AbstractScrollAreaMixin",
    "AbstractSlider",
    "AbstractSliderMixin",
    "AbstractSpinBox",
    "AbstractSpinBoxMixin",
    "Application",
    "BoxLayout",
    "ButtonGroup",
    "CalendarWidget",
    "CheckBox",
    "ColorDialog",
    "ColumnView",
    "ComboBox",
    "ComboBoxMixin",
    "CommandLinkButton",
    "CommonStyle",
    "CommonStyleMixin",
    "Completer",
    "DataWidgetMapper",
    "DateEdit",
    "DateTimeEdit",
    "DateTimeEditMixin",
    "Dial",
    "Dialog",
    "DialogButtonBox",
    "DialogMixin",
    "DockWidget",
    "DoubleSpinBox",
    "ErrorMessage",
    "FileDialog",
    "FileIconProvider",
    "FileSystemModel",
    "FocusFrame",
    "FontComboBox",
    "FontDialog",
    "FormLayout",
    "Frame",
    "FrameMixin",
    "Gesture",
    "GestureMixin",
    "GraphicsAnchorLayout",
    "GraphicsBlurEffect",
    "GraphicsColorizeEffect",
    "GraphicsDropShadowEffect",
    "GraphicsEffect",
    "GraphicsEffectMixin",
    "GraphicsEllipseItem",
    "GraphicsGridLayout",
    "GraphicsItem",
    "GraphicsItemGroup",
    "GraphicsItemMixin",
    "GraphicsLayout",
    "GraphicsLayoutItem",
    "GraphicsLayoutItemMixin",
    "GraphicsLayoutMixin",
    "GraphicsLineItem",
    "GraphicsLinearLayout",
    "GraphicsObject",
    "GraphicsObjectMixin",
    "GraphicsOpacityEffect",
    "GraphicsPathItem",
    "GraphicsPixmapItem",
    "GraphicsPolygonItem",
    "GraphicsProxyWidget",
    "GraphicsRectItem",
    "GraphicsRotation",
    "GraphicsScale",
    "GraphicsScene",
    "GraphicsSimpleTextItem",
    "GraphicsTextItem",
    "GraphicsTransform",
    "GraphicsTransformMixin",
    "GraphicsView",
    "GraphicsViewMixin",
    "GraphicsWidget",
    "GraphicsWidget",
    "GraphicsWidgetMixin",
    "GridLayout",
    "GroupBox",
    "HBoxLayout",
    "HeaderView",
    "InputDialog",
    "ItemDelegate",
    "ItemEditorCreatorBase",
    "ItemEditorFactory",
    "KeySequenceEdit",
    "LCDNumber",
    "Label",
    "Layout",
    "LayoutItem",
    "LayoutItemMixin",
    "LayoutMixin",
    "LineEdit",
    "ListView",
    "ListViewMixin",
    "ListWidget",
    "ListWidgetItem",
    "MainWindow",
    "MdiArea",
    "MdiSubWindow",
    "Menu",
    "MenuBar",
    "MessageBox",
    "PanGesture",
    "PinchGesture",
    # "KeyEventTransition",
    # "MouseEventTransition",
    # "GraphicsSceneHoverEvent",
    # "GraphicsSceneMouseEvent",
    # "GraphicsSceneWheelEvent",
    # "GraphicsSceneContextMenuEvent",
    # "GraphicsSceneDragDropEvent",
    # "GraphicsSceneHelpEvent",
    # "GraphicsSceneMoveEvent",
    # "GraphicsSceneResizeEvent",
    # "GraphicsSceneEvent",
    # "GestureEvent",
    "PlainTextDocumentLayout",
    "PlainTextEdit",
    "PlainTextEditMixin",
    "ProgressBar",
    "ProgressDialog",
    "ProxyStyle",
    "PushButton",
    "PushButtonMixin",
    "RadioButton",
    "RubberBand",
    "ScrollArea",
    "ScrollBar",
    "Scroller",
    "ScrollerProperties",
    "SizeGrip",
    "SizePolicy",
    "Slider",
    "SpacerItem",
    "SpinBox",
    "SplashScreen",
    "Splitter",
    "SplitterHandle",
    "StackedLayout",
    "StackedWidget",
    "StatusBar",
    "Style",
    "StyleFactory",
    "StyleMixin",
    "StyleOption",
    "StyleOptionButton",
    "StyleOptionComboBox",
    "StyleOptionComplex",
    "StyleOptionComplexMixin",
    "StyleOptionDockWidget",
    "StyleOptionFocusRect",
    "StyleOptionFrame",
    "StyleOptionGraphicsItem",
    "StyleOptionGroupBox",
    "StyleOptionHeader",
    "StyleOptionMenuItem",
    "StyleOptionMixin",
    "StyleOptionProgressBar",
    "StyleOptionRubberBand",
    "StyleOptionSizeGrip",
    "StyleOptionSlider",
    "StyleOptionSpinBox",
    "StyleOptionTab",
    "StyleOptionTabBarBase",
    "StyleOptionTabWidgetFrame",
    "StyleOptionTitleBar",
    "StyleOptionToolBar",
    "StyleOptionToolBox",
    "StyleOptionToolButton",
    "StyleOptionViewItem",
    "StylePainter",
    "StyledItemDelegate",
    "SwipeGesture",
    "SystemTrayIcon",
    "TabBar",
    "TabWidget",
    "TableView",
    "TableViewMixin",
    "TableWidget",
    "TableWidgetItem",
    "TableWidgetSelectionRange",
    "TapAndHoldGesture",
    "TapGesture",
    "TextBrowser",
    "TextEdit",
    "TextEditMixin",
    "TimeEdit",
    "ToolBar",
    "ToolBox",
    "ToolButton",
    "ToolTip",
    "TreeView",
    "TreeViewMixin",
    "TreeWidget",
    "TreeWidgetItem",
    "TreeWidgetItemIterator",
    "UndoView",
    "VBoxLayout",
    "WhatsThis",
    "Widget",
    "WidgetAction",
    "WidgetItem",
    "WidgetMixin",
    "Wizard",
    "WizardPage",
    "app",
    "custom_widgets",
    "itemdelegates",
    "itemmodels",
    "validators",
]



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Literal

from prettyqt.qt import QtWidgets
from prettyqt.utils import bidict


if TYPE_CHECKING:
    from collections.abc import Iterator


mod = QtWidgets.QScrollerProperties

FrameRateStr = Literal["fps_60", "fps_30", "fps_20", "standard"]

FRAME_RATES: bidict[FrameRateStr, mod.FrameRates] = bidict(
    fps_60=mod.FrameRates.Fps60,
    fps_30=mod.FrameRates.Fps30,
    fps_20=mod.FrameRates.Fps20,
    standard=mod.FrameRates.Standard,
)

OvershootPolicyStr = Literal["when_scrollable", "always_off", "always_on"]

OVERSHOOT_POLICY: bidict[OvershootPolicyStr, mod.OvershootPolicy] = bidict(
    when_scrollable=mod.OvershootPolicy.OvershootWhenScrollable,
    always_off=mod.OvershootPolicy.OvershootAlwaysOff,
    always_on=mod.OvershootPolicy.OvershootAlwaysOn,
)

ScrollmetricStr = Literal[
    "mouse_press_event_delay",
    "drag_start_distance",
    "drag_velocity_smoothing_factor",
    "axis_lock_threshold",
    "scrolling_curve",
    "deceleration_factor",
    "minimum_velocity",
    "maximum_velocity",
    "maximum_click_through_velocity",
    "accelerating_flick_maximum_time",
    "accelerating_flick_speedup_factor",
    "snap_position_ratio",
    "snap_time",
    "overshoot_drag_resistance_factor",
    "overshoot_drag_distance_factor",
    "overshoot_scroll_distance_factor",
    "overshoot_scroll_time",
    "horizontal_overshoot_policy",
    "vertical_overshoot_policy",
    "frame_rate",
    "scroll_metric_count",
]

SCROLL_METRIC: bidict[ScrollmetricStr, mod.ScrollMetric] = bidict(
    mouse_press_event_delay=mod.ScrollMetric.MousePressEventDelay,
    drag_start_distance=mod.ScrollMetric.DragStartDistance,
    drag_velocity_smoothing_factor=mod.ScrollMetric.DragVelocitySmoothingFactor,
    axis_lock_threshold=mod.ScrollMetric.AxisLockThreshold,
    scrolling_curve=mod.ScrollMetric.ScrollingCurve,
    deceleration_factor=mod.ScrollMetric.DecelerationFactor,
    minimum_velocity=mod.ScrollMetric.MinimumVelocity,
    maximum_velocity=mod.ScrollMetric.MaximumVelocity,
    maximum_click_through_velocity=mod.ScrollMetric.MaximumClickThroughVelocity,
    accelerating_flick_maximum_time=mod.ScrollMetric.AcceleratingFlickMaximumTime,
    accelerating_flick_speedup_factor=mod.ScrollMetric.AcceleratingFlickSpeedupFactor,
    snap_position_ratio=mod.ScrollMetric.SnapPositionRatio,
    snap_time=mod.ScrollMetric.SnapTime,
    overshoot_drag_resistance_factor=mod.ScrollMetric.OvershootDragResistanceFactor,
    overshoot_drag_distance_factor=mod.ScrollMetric.OvershootDragDistanceFactor,
    overshoot_scroll_distance_factor=mod.ScrollMetric.OvershootScrollDistanceFactor,
    overshoot_scroll_time=mod.ScrollMetric.OvershootScrollTime,
    horizontal_overshoot_policy=mod.ScrollMetric.HorizontalOvershootPolicy,
    vertical_overshoot_policy=mod.ScrollMetric.VerticalOvershootPolicy,
    frame_rate=mod.ScrollMetric.FrameRate,
    scroll_metric_count=mod.ScrollMetric.ScrollMetricCount,
)


class ScrollerProperties(QtWidgets.QScrollerProperties):
    """Stores the settings for a QScroller."""

    def __getitem__(self, metric: ScrollmetricStr | mod.ScrollMetric):
        return self.get_scroll_metric(metric)

    def __setitem__(self, metric: ScrollmetricStr | mod.ScrollMetric, value: Any):
        self.set_scroll_metric(metric, value)

    def keys(self) -> ScrollmetricStr:
        return SCROLL_METRIC.keys()

    def __iter__(self) -> Iterator[ScrollmetricStr]:
        return iter(SCROLL_METRIC.keys())

    def set_scroll_metric(self, metric: ScrollmetricStr | mod.ScrollMetric, value: Any):
        """Set scroll metric.

        Args:
            metric: Scroll metric to set
            value: Value to set
        """
        self.setScrollMetric(SCROLL_METRIC.get_enum_value(metric), value)

    def get_scroll_metric(self, metric: ScrollmetricStr | mod.ScrollMetric) -> Any:
        """Return scroll metric.

        Args:
            metric: Scroll metric to get

        Returns:
            state
        """
        return self.scrollMetric(SCROLL_METRIC.get_enum_value(metric))

    def get_scroll_metrics(self) -> dict[ScrollmetricStr, Any]:
        return {i: self.get_scroll_metric(i) for i in SCROLL_METRIC}


if __name__ == "__main__":
    props = ScrollerProperties()



"""Tests for `prettyqt` package."""

import pytest

import prettyqt.qt
from prettyqt.utils import InvalidParamError


@pytest.mark.skipif(prettyqt.qt.API.startswith("pyside"), reason="Only supported in PyQt")
def test_sciscintilla(qtbot):
    from prettyqt import gui, scintilla

    widget = scintilla.SciScintilla()
    widget.define_marker("circle", 0)
    with pytest.raises(InvalidParamError):
        widget.define_marker("test", 0)
    widget.set_marker_background_color("red", 0)
    widget.set_margins_background_color("green")
    widget.highlight_current_line(color="blue")
    widget.set_brace_matching("sloppy")
    with pytest.raises(InvalidParamError):
        widget.set_brace_matching("test")
    widget.set_text("test")
    assert widget.get_value() == "test"
    widget.set_syntaxhighlighter("python")
    widget.scroll_to_bottom()
    widget.append_text("test", newline=False)
    assert widget.text() == "testtest"
    widget.append_text("test", newline=True)
    assert widget.text() == "testtest\ntest"
    widget.set_font(gui.Font("Consolas"))
    widget.set_read_only()
    widget.set_value("test")



"""Tests for `prettyqt` package."""

import pathlib

import pytest

from prettyqt import core, qml
from prettyqt.utils import InvalidParamError


# def test_jsvalue():
#     val = qml.JSValue(2)
#     val["test"] = 1
#     assert val["test"].toInt() == 1
#     assert "test" in val
#     assert val.get_value() == 2


def test_jsengine():
    engine = qml.JSEngine()
    engine.install_extensions("translation")
    engine.eval("")


def test_qmlengine():
    engine = qml.QmlEngine()
    obj = core.Object()
    engine.set_object_ownership(obj, "javascript")
    with pytest.raises(InvalidParamError):
        engine.set_object_ownership(obj, "test")
    assert engine.get_object_ownership(obj) == "javascript"
    engine.add_plugin_path("")
    engine.add_import_path("")
    engine.get_plugin_paths()
    engine.get_import_paths()


def test_qmlapplicationengine(qtlog):
    with qtlog.disabled():
        engine = qml.QmlApplicationEngine()
        for _item in engine:
            pass
        path = pathlib.Path.cwd() / "tests" / "qmltest.qml"
        engine.load_data(path.read_text())


def test_qmlcomponent():
    comp = qml.QmlComponent()
    assert comp.get_status() == "null"
    # comp.load_url("", mode="asynchronous")
    comp.get_url()


def test_jsvalue():
    val = qml.JSValue(1)
    assert val.get_error_type() is None
    assert val.get_value() == 1
    repr(val)
    engine = qml.JSEngine()
    val = engine.new_array(2)
    val["test1"] = 1
    val["test2"] = 2
    assert val["test1"] == 1
    assert "test2" in val
    assert len(val) == 2
    del val["test2"]
    for _n, _v in val:
        pass
    val = qml.JSValue.from_object(None, engine)
    val = qml.JSValue.from_object(1, engine)
    val = qml.JSValue.from_object(["test"], engine)
    val = qml.JSValue.from_object(dict(a="b"), engine)



from __future__ import annotations

from prettyqt import gui
from prettyqt.utils import serializemixin


class Picture(serializemixin.SerializeMixin, gui.PaintDeviceMixin, gui.QPicture):
    """Paint device that records and replays QPainter commands."""



"""Create sound scenes in 3D space containing different sound sources."""

from __future__ import annotations

from prettyqt.qt.QtSpatialAudio import *  # noqa: F403

from .ambientsound import AmbientSound
from .audioengine import AudioEngine
from .audiolistener import AudioListener
from .audioroom import AudioRoom
from .spatialsound import SpatialSound
from prettyqt.qt import QtSpatialAudio

QT_MODULE = QtSpatialAudio

__all__ = ["AmbientSound", "AudioEngine", "AudioListener", "AudioRoom", "SpatialSound"]



from __future__ import annotations

import sys

from prettyqt import core, gui


class TextLengthValidator(gui.Validator):
    """Validator which checks whether given text has a specific length."""

    ID = "text_length"

    def __init__(
        self,
        minimum: int | None = None,
        maximum: int | None = None,
        parent: core.QObject | None = None,
    ):
        super().__init__(parent)
        self._minimum = 0 if minimum is None else minimum
        self._maximum = sys.maxsize if maximum is None else maximum

    def set_range(self, minimum: int | None, maximum: int | None):
        self._minimum = 0 if minimum is None else minimum
        self._maximum = sys.maxsize if maximum is None else maximum

    def get_range(self) -> tuple[int | None, int | None]:
        return (self._minimum, self._maximum)

    def set_minimum(self, minimum: int | None):
        self._minimum = 0 if minimum is None else minimum

    def get_minimum(self) -> int:
        return self._minimum

    def set_maximum(self, maximum: int | None):
        self._maximum = sys.maxsize if maximum is None else maximum

    def get_maximum(self) -> int:
        return self._maximum

    def validate(self, text: str, pos: int):
        if self._minimum <= len(text) <= self._maximum:
            return self.State.Acceptable, text, len(text)
        if len(text) <= self._maximum:
            return self.State.Intermediate, text, len(text)
        return self.State.Invalid, text, len(text)

    minimum = core.Property(
        int,
        get_minimum,
        set_minimum,
        doc="Minimum text length",
    )
    maximum = core.Property(
        int,
        get_maximum,
        set_maximum,
        doc="Maximum text length",
    )


if __name__ == "__main__":
    from prettyqt import widgets

    val = TextLengthValidator(maximum=10)
    app = widgets.app()
    widget = widgets.LineEdit("343")
    widget.setValidator(val)
    widget.show()
    app.exec()



from __future__ import annotations

from prettyqt import printsupport


class PrintDialog(printsupport.AbstractPrintDialogMixin, printsupport.QPrintDialog):
    """Dialog for specifying the printer's configuration."""

    @classmethod
    def setup_example(cls):
        return None


if __name__ == "__main__":
    from prettyqt import widgets

    app = widgets.app()
    dlg = PrintDialog()
    dlg.show()



from __future__ import annotations

from prettyqt import core, gui, widgets


class PreviewScrollBar(widgets.ScrollBar):
    """Scrollbar showing a small preview of a (Plain)TextEdit."""

    def __init__(self, *args, parent=None, **kwargs):
        super().__init__(*args, **kwargs)
        self._scrollarea = parent
        self._width = 150
        # self._line_margin = 152
        self._scale = 0.3
        # self.installEventFilter("slider_move_to_mouse_click")

    @classmethod
    def setup_example(cls):
        w = widgets.PlainTextEdit()
        return cls(parent=w)

    def sizeHint(self):
        expand = (
            core.QSize(self._width, 1)
            if self.is_vertical()
            else core.QSize(1, self._width)
        )
        return super().sizeHint().expandedTo(expand)

    def paintEvent(self, event):
        # viewport = self._scrollarea.viewport()
        doc = self._scrollarea.document()
        v_scroll = self._scrollarea.v_scrollbar
        super().paintEvent(event)
        with gui.Painter(self) as painter:
            painter.setRenderHint(painter.RenderHint.TextAntialiasing, True)

            # draw background
            painter.save()
            bg_color = gui.Palette().get_color("window")
            painter.setBrush(bg_color)
            painter.drawRect(self.geometry())
            painter.restore()

            # figure out positions
            painter.save()
            first_vis_line, last_vis_line = self._scrollarea.get_visible_line_span()
            # start_line = max(0, first_vis_line - self._line_margin)
            # end_line = min(last_vis_line + self._line_margin, doc.blockCount())
            # start_block = doc.findBlockByLineNumber(start_line)
            # end_block = doc.findBlockByLineNumber(end_line)
            opt = widgets.QStyleOptionSlider()
            opt.initFrom(self)
            opt.subControls = widgets.QStyle.SubControl.SC_None
            opt.activeSubControls = widgets.QStyle.SubControl.SC_None
            opt.orientation = self.orientation()
            opt.minimum = self.minimum()
            opt.maximum = self.maximum()
            opt.sliderPosition = self.sliderPosition()
            opt.sliderValue = self.value()
            opt.singleStep = self.singleStep()
            opt.pageStep = self.pageStep()
            gr = self.style().subControlRect(
                widgets.QStyle.ComplexControl.CC_ScrollBar,
                opt,
                widgets.QStyle.SubControl.SC_ScrollBarGroove,
                self,
            )
            textarea_height = 0
            b = doc.begin()
            while b != doc.end():
                r = self._scrollarea.blockBoundingRect(b)
                textarea_height += r.height()
                b = b.next()
            minimap_height = textarea_height * self._scale
            minimap_visible_height = min(minimap_height, v_scroll.height())
            # minimap_fully_visible = minimap_visible_height == minimap_height
            height_ratio = minimap_height / v_scroll.height()
            map_scroll_distance = minimap_height - minimap_visible_height
            if v_scroll.maximum():
                scroll_per_value = map_scroll_distance / v_scroll.maximum()
            else:
                scroll_per_value = 1
            # doc_x_margin = 1
            # doc_height = min(gr.height(), textarea_height)
            # yoffset = 1  # top instead of center-aligned (gr.height() - doc_height) / 2
            # doc_rect = core.QRect(
            #     core.QPoint(gr.left() + doc_x_margin, yoffset + gr.top()),
            #     QSize(gr.width() - doc_x_margin, doc_height),
            # )
            # self.map_grove_rect = doc_rect

            # max_ = max(self.maximum() + 1, 1)
            # visible_start = (
            #     self.value() * doc_height / (max_ + self.pageStep())
            #     + doc_rect.top()
            #     + 0.5
            # )
            # visible_end = (self.value() + self.pageStep()) * doc_height / (
            #     max_ + self.pageStep()
            # ) + doc_rect.top()
            # visible_rect = core.QRect(doc_rect)
            # visible_rect.moveTop(visible_start)
            # visible_rect.setHeight(visible_end - visible_start)
            # #  adjust the rectangles
            # slider_rect = self.style().subControlRect(QStyle.CC_ScrollBar, opt,
            # QStyle.SC_ScrollBarSlider, self)
            # slider_rect.setX(doc_x_margin)
            # slider_rect.setWidth(self.width() - doc_x_margin * 2)

            # if ((doc_height + 2 * doc_x_margin >= gr.height()) and
            # (slider_rect.height() > visible_rect.height() + 2))
            #     visible_rect.adjust(2, 0, -3, 0)
            #  else:
            #     visible_rect.adjust(1, 0, -1, 2)
            #     slider_rect.setTop(visible_rect.top() - 1)
            #     slider_rect.setBottom(visible_rect.bottom() + 1)

            painter.scale(self._scale, self._scale)
            offset = -(
                v_scroll.value() * scroll_per_value * (1 / self._scale)
            )  # height_ratio * (1/self._scale))
            # scale = self.height() / self._scrollarea.get_pixel_height()
            b = doc.begin()
            top = v_scroll.height() * (1 / self._scale)
            count = 0
            while b != doc.end():
                r = self._scrollarea.blockBoundingRect(b)
                if b.isVisible() and top >= offset >= 0:
                    layout = b.layout()
                    count += 1
                    layout.draw(painter, core.QPointF(0, offset))
                offset += r.height()
                b = b.next()
            painter.restore()
            # pos = v_scroll.value() / v_scroll.maximum()
            painter.setBrush(gui.QColor(120, 120, 120, 120))
            # slider_rect = self.style().subControlRect(
            #     widgets.QStyle.ComplexControl.CC_ScrollBar,
            #     opt,
            #     widgets.QStyle.SubControl.SC_ScrollBarSlider,
            #     self,
            # )
            handle_rect_height = (
                (last_vis_line - first_vis_line)
                / doc.blockCount()
                * gr.height()
                * height_ratio
            )

            handle_rect_y_pos = first_vis_line / doc.blockCount() * gr.height()
            handle_rect_y_pos *= v_scroll.height() / (
                v_scroll.height() + handle_rect_height
            )

            painter.drawRoundedRect(
                0, int(handle_rect_y_pos), self._width, int(handle_rect_height), 5, 5
            )

    def minimap_y_to_std_y(self, y: int) -> int:
        #  Check if the minimap fills the whole scrollbar
        if self.std_grove_rect.height() == self.map_grove_rect.height():
            return y
        #  check if y is on the step up/down
        if (y < self.std_grove_rect.top()) or (y > self.std_grove_rect.bottom()):
            return y
        if y < self.map_grove_rect.top():
            return self.std_grove_rect.top() + 1
        if y > self.map_grove_rect.bottom():
            return self.std_grove_rect.bottom() - 1
        #  check for div/0
        if self.map_grove_rect.height() == 0:
            return y
        newY = (
            (y - self.map_grove_rect.top())
            * self.std_grove_rect.height()
            / self.map_grove_rect.height()
        )
        newY += self.std_grove_rect.top()
        return newY


if __name__ == "__main__":
    app = widgets.app()
    app.set_style("fusion")
    widget = widgets.PlainTextEdit("\n".join(f"abc{i}" for i in range(1000)))
    widget.set_syntaxhighlighter("python")
    scrollbar = PreviewScrollBar("vertical", parent=widget)
    doc = widget.document()
    print(widget.viewport())
    print(widget.get_pixel_height())
    print(doc.pageSize())
    widget.setVerticalScrollBar(scrollbar)
    widget.show()
    with app.debug_mode():
        app.exec()



from __future__ import annotations

import logging
import multiprocessing
import os
import sys

import IPython
from qtconsole.client import QtKernelClient

from prettyqt import core, ipython, widgets


logger = logging.getLogger(__name__)

# ipython = IPython.get_ipython()
# ipython.magic(f"gui qt{qt.QT_VERSION}")

# asyncio.set_event_loop_policy(asyncio.WindowsSelectorEventLoopPolicy())


def run_server(connection_file: str | os.PathLike[str]):
    IPython.embed_kernel(
        local_ns=sys._getframe(1).f_locals,
        connection_file=os.fspath(connection_file),
        gui="qt6",
    )


class OutOfProcessIPythonWidget(ipython.BaseIPythonWidget):
    """Convenience class for a live IPython console widget running out-of-process."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        temp_path = core.Dir.get_temp_path()
        self.connection_file = temp_path / f"connection-{os.getpid()}.json"
        self.p = multiprocessing.Process(target=run_server, args=(self.connection_file,))
        self.p.daemon = True
        self.p.start()
        logger.info("waiting for connection file creation....")
        for _ in range(500):
            if self.connection_file.exists() and self.connection_file.stat().st_size > 0:
                break
            widgets.app().sleep(0.1)
        else:
            raise FileNotFoundError(self.connection_file)
        logger.info("Connection file found. Opening channels.")
        kernel_client = QtKernelClient(connection_file=os.fspath(self.connection_file))
        kernel_client.load_connection_file()  # can throw FileNotFoundError
        kernel_client.start_channels()
        self.kernel_client = kernel_client

    def shutdown(self):
        """Stop IPython server process and clean up."""
        logger.info("shutting down IPython MP kernel")
        assert self.kernel_client, "Kernel client not initialized"
        self.kernel_client.stop_channels()
        self.kernel_client.shutdown()
        self.p.terminate()
        self.connection_file.unlink(missing_ok=True)
        logger.info("shutdown successful.")


if __name__ == "__main__":
    from prettyqt.custom_widgets import commandpalette

    app = widgets.app()
    cp = commandpalette.CommandPalette()
    app.set_style("fusion")
    console_widget = OutOfProcessIPythonWidget(app)
    console_widget.print_text("hallo")
    console_widget.show()
    console_widget.add_shortcut("Ctrl+P", callback=cp.show)
    cp.populate_from_widget(console_widget)
    app.exec()



from __future__ import annotations

from typing import Literal

from prettyqt import widgets
from prettyqt.utils import bidict


DirectionStr = Literal["left_to_right", "right_to_left", "top_to_bottom", "bottom_to_top"]

DIRECTION: bidict[DirectionStr, widgets.QBoxLayout.Direction] = bidict(
    left_to_right=widgets.QBoxLayout.Direction.LeftToRight,
    right_to_left=widgets.QBoxLayout.Direction.RightToLeft,
    top_to_bottom=widgets.QBoxLayout.Direction.TopToBottom,
    bottom_to_top=widgets.QBoxLayout.Direction.BottomToTop,
)


class BoxLayoutMixin(widgets.LayoutMixin):
    # def __init__(
    #     self,
    #     orientation: Literal["horizontal", "vertical"] = "horizontal",
    #     parent: widgets.QWidget | None = None,
    #     margin: int | None = None,
    # ):

    #     if margin is not None:
    #         self.set_margin(margin)

    # def __setstate__(self, state):
    #     super().__setstate__(state)
    #     self.set_direction(state["direction"])
    #     for item in state["items"]:
    #         self.add(item)

    # def __reduce__(self):
    #     return type(self), (), self.__getstate__()

    # def __add__(self, other: widgets.QWidget | widgets.QLayout):
    #     self.add(other)
    #     return self

    # def add(self, *item):
    #     for i in item:
    #         if isinstance(i, widgets.QWidget):
    #             self.addWidget(i)
    #         else:
    #             self.addLayout(i)

    def add_stretch(self, stretch: int = 0):
        self.addStretch(stretch)

    def add_spacing(self, size: int):
        self.addSpacing(size)

    def set_direction(self, direction: DirectionStr | BoxLayout.Direction):
        """Set the direction.

        Args:
            direction: direction
        """
        self.setDirection(DIRECTION.get_enum_value(direction))

    def get_direction(self) -> DirectionStr:
        """Return current direction.

        Returns:
            direction
        """
        return DIRECTION.inverse[self.direction()]


class BoxLayout(BoxLayoutMixin, widgets.QBoxLayout):
    """Lines up child widgets horizontally or vertically."""

    ID = "box"


if __name__ == "__main__":
    app = widgets.app()
    layout = BoxLayout("vertical")
    widget = widgets.Widget()
    widget2 = widgets.RadioButton("Test")
    layout.add(widget2)
    widget.set_layout(layout)
    widget.show()
    app.exec()



from __future__ import annotations

from typing import Literal

from prettyqt import gui
from prettyqt.qt import QtPrintSupport
from prettyqt.utils import bidict


COLOR_MODE = bidict(
    color=QtPrintSupport.QPrinter.ColorMode.Color,
    gray_scale=QtPrintSupport.QPrinter.ColorMode.GrayScale,
)

ColorModeStr = Literal["color", "gray_scale"]

DUPLEX_MODE = bidict(
    none=QtPrintSupport.QPrinter.DuplexMode.DuplexNone,
    auto=QtPrintSupport.QPrinter.DuplexMode.DuplexAuto,
    long_side=QtPrintSupport.QPrinter.DuplexMode.DuplexLongSide,
    short_side=QtPrintSupport.QPrinter.DuplexMode.DuplexShortSide,
)

DuplexModeStr = Literal["none", "auto", "long_side", "short_side"]

OUTPUT_FORMAT = bidict(
    native=QtPrintSupport.QPrinter.OutputFormat.NativeFormat,
    pdf=QtPrintSupport.QPrinter.OutputFormat.PdfFormat,
)

OutputFormatStr = Literal["native", "pdf"]

PAGE_ORDER = bidict(
    first_page_last=QtPrintSupport.QPrinter.PageOrder.FirstPageFirst,
    last_page_first=QtPrintSupport.QPrinter.PageOrder.LastPageFirst,
)

PageOrderStr = Literal["first_page_last", "last_page_first"]

PRINT_RANGE = bidict(
    all_pages=QtPrintSupport.QPrinter.PrintRange.AllPages,
    selection=QtPrintSupport.QPrinter.PrintRange.Selection,
    page_range=QtPrintSupport.QPrinter.PrintRange.PageRange,
    current_page=QtPrintSupport.QPrinter.PrintRange.CurrentPage,
)

PrintRangeStr = Literal["all_pages", "selection", "page_range", "current_page"]

PRINTER_MODE = bidict(
    screen_resolution=QtPrintSupport.QPrinter.PrinterMode.ScreenResolution,
    printer_resolution=QtPrintSupport.QPrinter.PrinterMode.PrinterResolution,
    high_resolution=QtPrintSupport.QPrinter.PrinterMode.HighResolution,
)

PrinterModeStr = Literal["screen_resolution", "printer_resolution", "high_resolution"]

PRINTER_STATE = bidict(
    idle=QtPrintSupport.QPrinter.PrinterState.Idle,
    active=QtPrintSupport.QPrinter.PrinterState.Active,
    aborted=QtPrintSupport.QPrinter.PrinterState.Aborted,
    error=QtPrintSupport.QPrinter.PrinterState.Error,
)

PrinterStateStr = Literal["idle", "active", "aborted", "error"]


class Printer(gui.PagedPaintDeviceMixin, QtPrintSupport.QPrinter):
    """Paint device that paints on a printer."""

    # def get_source(self) -> pathlib.Path:
    #     return pathlib.Path(self.source().toLocalFile())

    def get_duplex(self) -> DuplexModeStr:
        return DUPLEX_MODE.inverse[self.duplex()]

    def get_pdf_version(self) -> gui.pagedpaintdevice.PdfVersionStr:
        return gui.pagedpaintdevice.PDF_VERSION.inverse[self.pdfVersion()]

    def set_pdf_version(
        self,
        version: gui.pagedpaintdevice.PdfVersionStr | gui.PagedPaintDevice.PdfVersion,
    ):
        """Set pdf version.

        Args:
            version: pdf version
        """
        self.setPdfVersion(gui.pagedpaintdevice.PDF_VERSION.get_enum_value(version))



from __future__ import annotations

from typing import Literal

from prettyqt import core
from prettyqt.utils import get_repr


class MetaProperty:
    """Meta-data about a property."""

    def __init__(self, metaproperty: core.QMetaProperty):
        self.item = metaproperty

    def __bool__(self):
        return self.item.isValid()

    def __repr__(self):
        return get_repr(self, self.get_name())

    def __getattr__(self, val):
        return getattr(self.item, val)

    def get_name(self) -> str:
        return self.item.name()  # type: ignore

    def get_meta_type(self) -> core.MetaType:
        return core.MetaType(self.userType())  # same as self.metaType().id()

    def get_python_type(self) -> type:
        return self.get_meta_type().get_type()

    def get_notify_signal(self) -> core.MetaMethod | None:
        if (signal := self.notifySignal()).isValid():
            return core.MetaMethod(signal)
        return None

    def get_enumerator(self) -> core.MetaEnum | None:
        if (enumerator := self.enumerator()).isValid():
            return core.MetaEnum(enumerator)
        return None

    def get_enumerator_type(self) -> Literal["flag", "enum"] | None:
        if self.isFlagType():
            return "flag"
        if self.isEnumType():
            return "enum"
        return None


if __name__ == "__main__":
    from prettyqt import widgets

    app = widgets.app()
    widget = widgets.Widget()
    metaobj = widget.get_metaobject()
    prop = metaobj.get_property("x")
    value = prop.read(widget)



from __future__ import annotations

from prettyqt.qt import QtCore
from prettyqt.utils import get_repr


class MimeType(QtCore.QMimeType):
    """Describes types of file or data, represented by a MIME type string."""

    def __bool__(self):
        return self.isValid()

    def __repr__(self):
        return get_repr(self, self.name())

    def __str__(self):
        return self.name()



from __future__ import annotations

from typing import Literal

from prettyqt.qt import QtCore
from prettyqt.utils import bidict


PriorityStr = Literal[
    "idle", "lowest", "low", "normal", "high", "highest", "time_critical", "inherit"
]

PRIORITY: bidict[PriorityStr, QtCore.QThread.Priority] = bidict(
    idle=QtCore.QThread.Priority.IdlePriority,
    lowest=QtCore.QThread.Priority.LowestPriority,
    low=QtCore.QThread.Priority.LowPriority,
    normal=QtCore.QThread.Priority.NormalPriority,
    high=QtCore.QThread.Priority.HighPriority,
    highest=QtCore.QThread.Priority.HighestPriority,
    time_critical=QtCore.QThread.Priority.TimeCriticalPriority,
    inherit=QtCore.QThread.Priority.InheritPriority,
)


class Thread(QtCore.QThread):
    """Platform-independent way to manage threads."""

    def get_priority(self) -> PriorityStr:
        return PRIORITY.inverse[self.priority()]



from __future__ import annotations

from prettyqt import core, widgets
from prettyqt.qt import QtGui


class StatusBar(widgets.WidgetMixin, widgets.QStatusBar):
    """Horizontal bar suitable for presenting status information."""

    def __add__(self, other: QtGui.QAction | widgets.QWidget) -> StatusBar:
        match other:
            case QtGui.QAction():
                self.addAction(other)
                return self
            case widgets.QWidget():
                self.addWidget(other)
                return self
            case _:
                raise TypeError(other)

    def add_threadpool_info(self, threadpool: core.ThreadPool):
        class ThreadPoolLabel(widgets.Label):
            def update_job_count(self, num_jobs: int):
                text = f"Running jobs: {num_jobs}" if num_jobs > 0 else "No running jobs"
                self.set_text(text)

        status_label = ThreadPoolLabel(self)
        threadpool.job_num_updated.connect(status_label.update_job_count)
        status_label.update_job_count(threadpool.activeThreadCount())
        progress_bar = widgets.ProgressBar(self, text_visible=False)
        progress_bar.hide()
        progress_bar.setRange(0, 0)
        progress_bar.setFixedSize(200, 20)
        threadpool.busy_state_changed.connect(progress_bar.setVisible)
        self.addPermanentWidget(status_label)
        self.addPermanentWidget(progress_bar)

    def add_widget(self, widget: widgets.QWidget, permanent: bool = False) -> None:
        if permanent:
            self.addPermanentWidget(widget)
        else:
            self.addWidget(widget)

    def show_message(self, message: str, timeout: int = 0) -> None:
        self.showMessage(message, timeout)


if __name__ == "__main__":
    app = widgets.app()
    dlg = widgets.MainWindow()
    status_bar = StatusBar()
    status_bar.set_color("black")
    label = widgets.Label("test")
    status_bar.addWidget(label)
    status_bar.setup_default_bar()
    dlg.setStatusBar(status_bar)
    dlg.show()
    app.exec()



from __future__ import annotations

from typing import Literal

from prettyqt import core, gui, quick
from prettyqt.utils import bidict


mod = quick.QQuickPaintedItem

PerformanceHintStr = Literal["fast_fbo_resizing"]

PERFORMANCE_HINT: bidict[PerformanceHintStr, mod.PerformanceHint] = bidict(
    fast_fbo_resizing=mod.PerformanceHint.FastFBOResizing,
)

RenderTargetStr = Literal["image", "framebuffer_object", "inverted_y_framebuffer_object"]

RENDER_TARGET: bidict[RenderTargetStr, mod.RenderTarget] = bidict(
    image=mod.RenderTarget.Image,
    framebuffer_object=mod.RenderTarget.FramebufferObject,
    inverted_y_framebuffer_object=mod.RenderTarget.InvertedYFramebufferObject,
)


class QuickPaintedItem(quick.QuickItemMixin, quick.QQuickPaintedItem):
    """Way to use the QPainter API in the QML Scene Graph."""

    def get_fill_color(self) -> gui.Color:
        return gui.Color(self.fillColor())

    def get_texture_size(self) -> core.Size:
        return core.Size(self.textureSize())

    def set_render_target(self, target: RenderTargetStr | mod.RenderTarget):
        """Set the render target.

        Args:
            target: render target to use
        """
        self.setRenderTarget(RENDER_TARGET[target])

    def get_render_target(self) -> RenderTargetStr:
        """Return the render target.

        Returns:
            render target
        """
        return RENDER_TARGET.inverse[self.renderTarget()]



from __future__ import annotations

import contextlib
import sys
from typing import TYPE_CHECKING, SupportsInt

from prettyqt import constants, core, gui, iconprovider
from prettyqt.utils import colors, datatypes


if TYPE_CHECKING:
    from collections.abc import Iterator


class GuiApplicationMixin(core.CoreApplicationMixin):
    palette_changed = core.Signal(gui.Palette)

    def __init__(self, *args, **kwargs):
        self._window_icon_color = "darkcyan"
        super().__init__(*args, **kwargs)

    def event(self, e):
        match e.type():
            case core.QEvent.Type.ApplicationPaletteChange:
                self.palette_changed.emit(gui.Palette(self.palette()))
        return super().event(e)

    @classmethod
    @contextlib.contextmanager
    def override_cursor(cls, cursor: constants.CursorShapeStr):
        cls.set_override_cursor(cursor)
        yield cursor
        cls.restore_override_cursor()

    @classmethod
    def set_override_cursor(cls, cursor: constants.CursorShapeStr):
        crs = gui.Cursor(constants.CURSOR_SHAPE[cursor])
        cls.setOverrideCursor(crs)

    @classmethod
    def restore_override_cursor(cls):
        cls.restoreOverrideCursor()

    @classmethod
    def get_clipboard(cls) -> gui.Clipboard:
        return gui.Clipboard(cls.clipboard())

    @classmethod
    @contextlib.contextmanager
    def edit_palette(cls) -> Iterator[gui.Palette]:
        palette = gui.Palette(cls.palette())
        yield palette
        cls.setPalette(palette)

    @classmethod
    def find_window(cls, h_wnd: SupportsInt) -> gui.QWindow:
        for window in cls.topLevelWindows():
            if window and int(window.winId()) == int(h_wnd):
                return window
        return None

    def set_layout_direction(
        self, direction: constants.LayoutDirectionStr | constants.LayoutDirection
    ):
        """Set layout direction.

        Args:
            direction: layout direction
        """
        self.setLayoutDirection(constants.LAYOUT_DIRECTION.get_enum_value(direction))

    def get_layout_direction(self) -> constants.LayoutDirectionStr:
        """Get the current layout direction.

        Returns:
            layout direction
        """
        return constants.LAYOUT_DIRECTION.inverse[self.layoutDirection()]

    @classmethod
    def set_high_dpi_scale_factor_rounding_policy(
        cls,
        policy: (
            constants.HighDpiScaleFactorRoundingPolicyStr
            | constants.HighDpiScaleFactorRoundingPolicy
        ),
    ):
        """Set high dpi scale factor rounding policy.

        Args:
            policy: rounding policy
        """
        cls.setHighDpiScaleFactorRoundingPolicy(
            constants.HIGH_DPI_SCALE_FACTOR_ROUNDING_POLICY.get_enum_value(policy)
        )

    @classmethod
    def get_high_dpi_scale_factor_rounding_policy(
        cls,
    ) -> constants.HighDpiScaleFactorRoundingPolicyStr:
        """Get the current high dpi scale factor rounding policy.

        Returns:
            rounding policy
        """
        return constants.HIGH_DPI_SCALE_FACTOR_ROUNDING_POLICY.inverse[
            cls.highDpiScaleFactorRoundingPolicy()
        ]

    @classmethod
    def get_application_state(cls) -> list[constants.ApplicationStateStr]:
        """Get the current application state.

        Returns:
            application state
        """
        return [
            k
            for k, v in constants.APPLICATION_STATES.items()
            if v & cls.applicationState()  # type: ignore
        ]

    def get_primary_screen(self) -> gui.Screen:
        return gui.Screen(self.primaryScreen())

    def get_screen_at(self, point: datatypes.PointType) -> gui.Screen:
        p = datatypes.to_point(point)
        return gui.Screen(self.screenAt(p))

    def get_screens(self) -> list[gui.Screen]:
        return [gui.Screen(i) for i in self.screens()]

    @classmethod
    def get_input_method(cls) -> gui.InputMethod:
        return gui.InputMethod(cls.inputMethod())

    @classmethod
    def copy_to_clipboard(cls, text: str):
        """Sets clipboard to supplied text."""
        cb = cls.clipboard()
        cb.clear(mode=cb.Mode.Clipboard)
        cb.setText(text, mode=cb.Mode.Clipboard)

    @classmethod
    def get_font(cls) -> gui.Font:
        return gui.Font(cls.font())

    def set_icon(self, icon: datatypes.IconType):
        """Set the default window icon.

        Args:
            icon: icon to use
        """
        color = self.get_window_icon_color()
        icon = iconprovider.get_icon(icon, color=color)
        self.setWindowIcon(icon)

    def get_icon(self) -> gui.Icon | None:
        icon = self.windowIcon()
        return None if icon.isNull() else gui.Icon(self.windowIcon())

    @classmethod
    def set_palette(cls, palette: constants.ThemeStr | gui.QPalette):
        match palette:
            case "default":
                pal = gui.Palette()
            case "dark":
                pal = gui.Palette.create_dark_palette()
            case _:
                pal = palette
        cls.setPalette(pal)

    @classmethod
    def get_keyboard_modifiers(cls) -> list[constants.KeyboardModifierStr]:
        return constants.KEYBOARD_MODIFIERS.get_list(cls.keyboardModifiers())

    @classmethod
    def query_keyboard_modifiers(cls) -> list[constants.KeyboardModifierStr]:
        return constants.KEYBOARD_MODIFIERS.get_list(cls.queryKeyboardModifiers())

    @classmethod
    def get_palette(cls) -> gui.Palette:
        return gui.Palette(cls.palette())

    def set_badge_number(self, number: int | None):
        self.setBadgeNumber(number or 0)

    def set_progress_value(self, value: int, total: int = 100):
        windows = self.topLevelWindows()
        if not windows:
            return
        if sys.platform.startswith("win"):
            from prettyqt.utils.platforms.windows import taskbaritem

            window_id = windows[0].winId()
            tb = taskbaritem.TaskBarItem(window_id)
            tb.set_progress_value(value, total)

    def set_window_icon_color(self, color: datatypes.ColorType):
        self._window_icon_color = colors.get_color(color).name()

    def get_window_icon_color(self):
        return self._window_icon_color

    window_icon_color = core.Property(
        str,
        get_window_icon_color,
        set_window_icon_color,
        doc="Global window icon color",
    )


class GuiApplication(GuiApplicationMixin, gui.QGuiApplication):
    """Manages the GUI application's control flow and main settings."""


if __name__ == "__main__":
    app = gui.app()
    app.set_badge_number(5)
    app.exec()



"""Classes and functions for rendering PDF documents."""

from __future__ import annotations

from prettyqt.qt.QtPdf import *  # noqa: F403

from .pdfbookmarkmodel import PdfBookmarkModel
from .pdfdocument import PdfDocument
from .pdfsearchmodel import PdfSearchModel
from prettyqt.qt import QtPdf

QT_MODULE = QtPdf


__all__ = ["PdfBookmarkModel", "PdfDocument", "PdfSearchModel"]



"""Tests for `prettyqt` package."""

import importlib.util
import itertools
import pathlib
import pickle
import sys

import pytest

from prettyqt import constants, core, gui, widgets
from prettyqt.utils import InvalidParamError


def pickle_roundtrip(something):
    path = pathlib.Path("data.pkl")
    with path.open("wb") as jar:
        pickle.dump(something, jar)
    with path.open("rb") as jar:
        return pickle.load(jar)


def test_action(qtbot):
    action = gui.Action()
    action.set_tooltip("test")
    action.set_enabled()
    action.set_disabled()
    action.set_icon(None)
    action.set_icon("mdi.timer")
    action.set_shortcut("Ctrl+A")
    assert action.shortcut().toString() == "Ctrl+A"
    assert action.toolTip() == "test"
    action.set_priority("low")
    with pytest.raises(InvalidParamError):
        action.set_priority("test")
    assert action.get_priority() == "low"
    action.set_menu_role("preferences")
    with pytest.raises(InvalidParamError):
        action.set_menu_role("test")
    assert action.get_menu_role() == "preferences"
    action.set_shortcut_context("widget_with_children")
    with pytest.raises(InvalidParamError):
        action.set_shortcut_context("test")
    assert action.get_shortcut_context() == "widget_with_children"
    action.show_shortcut_in_contextmenu()
    action.set_menu(widgets.Menu())


def test_actiongroup(qtbot):
    group = gui.ActionGroup(None)
    group.set_exclusion_policy(None)
    group.set_exclusion_policy("exclusive")
    with pytest.raises(InvalidParamError):
        group.set_exclusion_policy("test")
    assert group.get_exclusion_policy() == "exclusive"
    act = gui.Action()
    group.addAction(act)
    assert group[0] == act
    assert act in group
    assert len(group) == 1


def test_brush():
    brush = gui.Brush()
    bytes(brush)
    brush.set_style("cross")
    with pytest.raises(InvalidParamError):
        brush.set_style("test")
    assert brush.get_style() == "cross"


@pytest.mark.skipif(sys.platform == "darwin", reason="Somehow fails on OSX")
def test_clipboard(qapp):
    mimedata = core.QMimeData()
    pixmap = gui.Pixmap(100, 100)
    image = pixmap.toImage()
    cb = qapp.get_clipboard()
    cb.set_mimedata(mimedata)
    cb.set_image(image)
    cb.set_pixmap(pixmap)
    assert cb.get_mimedata()
    assert cb.get_image() == image
    assert cb.get_pixmap().size() == pixmap.size()


def test_color(qapp):
    color = gui.Color()
    assert color.get_spec() == "invalid"
    color.set_color("gray")
    color = pickle_roundtrip(color)
    assert str(color) == "#808080"
    color.convert_to("rgb")
    color.convert_to("hsv")
    color.convert_to("cmyk")
    color.convert_to("hsl")
    color.convert_to("extended_rgb")
    # color.as_qt()


def test_colorspace():
    space = gui.ColorSpace()
    bytes(space)
    assert not bool(space)
    space = pickle_roundtrip(space)
    space.set_primaries("dci_p3_d65")
    with pytest.raises(InvalidParamError):
        space.set_primaries("test")
    assert space.get_primaries() == "dci_p3_d65"
    space.set_transfer_function("pro_photo_rgb")
    with pytest.raises(InvalidParamError):
        space.set_transfer_function("test")
    assert space.get_transfer_function() == "pro_photo_rgb"


def test_cursor(qapp):
    cursor = gui.Cursor()
    cursor.set_shape("arrow")
    with pytest.raises(InvalidParamError):
        cursor.set_shape("test")
    assert cursor.get_shape() == "arrow"
    cursor = pickle_roundtrip(cursor)
    bytes(cursor)
    cursor.get_pos()


# def test_desktopservices():
#     gui.DesktopServices.open_url("test")


def test_doublevalidator():
    val = gui.DoubleValidator()
    val.setRange(0, 9)
    val = pickle_roundtrip(val)
    assert val.is_valid_value("4")
    assert not val.is_valid_value("10")


def test_font(qapp):
    font = gui.Font("Consolas")
    assert font.metrics
    font = gui.Font.mono()
    with pytest.raises(InvalidParamError):
        font.set_style_hint("test")
    font.set_style_hint("monospace")
    font.set_weight("thin")
    with pytest.raises(InvalidParamError):
        font.set_weight("test")
    assert font.get_weight() == "thin"
    font.set_capitalization("small_caps")
    with pytest.raises(InvalidParamError):
        font.set_capitalization("test")
    assert font.get_capitalization() == "small_caps"
    font.set_style("oblique")
    with pytest.raises(InvalidParamError):
        font.set_style("test")
    assert font.get_style() == "oblique"
    font.set_hinting_preference("vertical")
    with pytest.raises(InvalidParamError):
        font.set_hinting_preference("test")
    assert font.get_hinting_preference() == "vertical"
    font.set_letter_spacing("absolute", 20)
    with pytest.raises(InvalidParamError):
        font.set_letter_spacing("test", 20)
    assert font.get_letter_spacing_type() == "absolute"


def test_fontdatabase():
    p = pathlib.Path()
    gui.FontDatabase.add_fonts_from_folder(p)
    gui.FontDatabase.get_system_font("smallest_readable")
    with pytest.raises(InvalidParamError):
        gui.FontDatabase.get_system_font("test")


def test_fontinfo():
    font = gui.Font("Consolas")
    fontinfo = gui.FontInfo(font)
    assert fontinfo.get_style_hint() == "any"


def test_fontmetrics():
    font = gui.Font("Consolas")
    fontmetrics = gui.FontMetrics(font)
    val = fontmetrics.elided_text("This is a test", mode="right", width=40)
    with pytest.raises(InvalidParamError):
        val = fontmetrics.elided_text("This is a test", mode="test", width=40)
    assert len(val) < 10
    fontmetrics.get_bounding_rect("test")
    fontmetrics.get_tight_bounding_rect("test")


def test_fontmetricsf():
    font = gui.Font("Consolas")
    fontmetrics = gui.FontMetricsF(font)
    val = fontmetrics.elided_text("This is a test", mode="right", width=40)
    with pytest.raises(InvalidParamError):
        val = fontmetrics.elided_text("This is a test", mode="test", width=40)
    assert len(val) < 10
    fontmetrics.get_bounding_rect("test")
    fontmetrics.get_tight_bounding_rect("test")


def test_gradient():
    grad = gui.Gradient()
    grad.set_coordinate_mode("object")
    assert grad.get_coordinate_mode() == "object"
    with pytest.raises(InvalidParamError):
        grad.set_coordinate_mode("test")
    grad.set_spread("repeat")
    assert grad.get_spread() == "repeat"
    with pytest.raises(InvalidParamError):
        grad.set_spread("test")
    assert grad.get_type() == "none"


def test_guiapplication():
    with gui.GuiApplication.override_cursor("forbidden"):
        pass


def test_icon(qapp):
    icon = gui.Icon()
    icon.for_color("black")
    icon = pickle_roundtrip(icon)
    with pytest.raises(InvalidParamError):
        icon.get_available_sizes(mode="test")
    with pytest.raises(InvalidParamError):
        icon.get_available_sizes(state="test")
    icon.get_available_sizes()
    icon.add_pixmap(b"a")
    icon.get_actual_size((256, 256))


def test_iconengine(qapp):
    engine = gui.IconEngine()
    engine.get_available_sizes()
    engine.get_actual_size((100, 100))
    engine.get_actual_size(100)
    with pytest.raises(InvalidParamError):
        engine.get_actual_size(100, state="test")
    with pytest.raises(InvalidParamError):
        engine.get_actual_size(100, mode="test")
    px = gui.Pixmap()
    engine.add_pixmap(px, mode="normal", state="off")


def test_image(qapp):
    img = gui.Image()
    img = pickle_roundtrip(img)
    bytes(img)


@pytest.mark.skipif(importlib.util.find_spec("PIL") is None, reason="PIL not installed")
def test_image_pil(qapp):
    image = gui.Pixmap(100, 100).to_image()
    pil_image = image.to_pil()
    assert gui.Image.from_pil(pil_image)


def test_imageiohandler():
    handler = gui.ImageIOHandler()
    handler.get_format()
    handler.set_option("gamma", 0.5)
    assert handler.get_option("gamma") is None
    handler["quality"] = 0.6
    assert handler["quality"] is None
    assert handler.supports_option("animation") is False


def test_imagereader():
    reader = gui.ImageReader()
    assert reader.get_error() == "unknown"
    reader.get_background_color()
    reader.get_clip_rect()
    reader.get_current_image_rect()
    reader.get_scaled_clip_rect()
    reader.get_size()
    reader.get_scaled_size()
    reader.get_subtype()
    reader.get_supported_subtypes()
    reader.set_format("xsn")
    assert reader.get_format() == "xsn"
    assert reader.get_transformation() == "none"
    reader.read_image()
    assert reader.supports_option("gamma") is False


def test_imagewriter(qapp):
    writer = gui.ImageWriter()
    writer["test"] = "Test"
    assert writer.get_error() == "unknown"
    writer.set_subtype("xsn")
    assert writer.get_subtype() == "xsn"
    writer.get_supported_image_formats()
    writer.get_supported_subtypes()
    writer.set_format("abc")
    assert writer.get_format() == "abc"
    writer.set_transformation("rotate_270")
    assert writer.get_transformation() == "rotate_270"


def test_inputmethod(qapp):
    input_method = qapp.get_input_method()
    input_method.get_anchor_rectangle()
    input_method.get_cursor_rectangle()
    input_method.get_input_item_clip_rectangle()
    input_method.get_input_item_rectangle()
    input_method.get_keyboard_rectangle()
    input_method.get_locale()
    assert input_method.get_input_direction() == "left_to_right"


def test_intvalidator():
    val = gui.IntValidator()
    val.setRange(0, 9)
    val = pickle_roundtrip(val)
    assert val.is_valid_value("4")
    assert not val.is_valid_value("10")


def test_keysequence():
    assert (
        gui.KeySequence.to_shortcut_str(
            constants.Key.Key_A, constants.KeyboardModifier.ShiftModifier
        )
        == "Shift+A"
    )
    comb = core.KeyCombination(
        constants.KeyboardModifier.ShiftModifier, constants.Key.Key_A
    )
    assert gui.KeySequence.to_shortcut_str(comb) == "Shift+A"
    seq = gui.KeySequence("Ctrl+C")
    assert seq.get_matches("Ctrl+C") == "exact"
    pickle_roundtrip(seq)


def test_movie():
    movie = gui.Movie()
    with pytest.raises(InvalidParamError):
        movie.set_cache_mode("test")
    movie.set_cache_mode("all")
    assert movie.get_cache_mode() == "all"
    movie = pickle_roundtrip(movie)
    repr(movie)


def test_screen(qapp):
    screen = qapp.get_primary_screen()
    screen.get_geometry()
    screen.get_size()
    screen.get_available_geometry()
    screen.get_available_size()
    screen.get_available_virtual_geometry()
    screen.get_available_virtual_size()
    screen.get_virtual_geometry()
    screen.get_virtual_size()
    screen.get_native_orientation()
    screen.get_orientation()
    screen.get_primary_orientation()
    screen.get_physical_size()
    screen.get_angle_between("primary", "landscape")
    w = widgets.Widget()
    screen.grab_window(w.winId())
    screen.get_virtual_siblings()


def test_shortcut(qtbot):
    w = widgets.Widget()
    qtbot.addWidget(w)
    seq = gui.KeySequence("Ctrl+C")
    shortcut = gui.Shortcut(seq, w)
    assert str(shortcut) == "Ctrl+C"
    shortcut.set_context("application")
    with pytest.raises(InvalidParamError):
        shortcut.set_context("test")
    assert shortcut.get_context() == "application"
    assert shortcut.get_key() == seq


# def test_sessionmanager():
#     manager = gui.SessionManager()
#     with pytest.raises(InvalidParamError):
#         manager.set_restart_hint("test")
#     manager.set_restart_hint("immediately")
#     assert manager.get_restart_hint() == "immediately"


def test_standarditem(qapp):
    item = gui.StandardItem()
    item = pickle_roundtrip(item)
    # item[constants.USER_ROLE] = "test"
    # assert item[constants.USER_ROLE] == "test"
    item.set_icon("mdi.timer")
    item.set_checkstate("unchecked")
    with pytest.raises(InvalidParamError):
        item.set_checkstate("test")
    assert item.get_checkstate() == "unchecked"
    item.get_background()
    item.get_foreground()
    item.get_font()
    item.get_icon()
    bytes(item)
    item.clone()
    item.add_item(
        "Test",
        icon="mdi.timer",
        data={1: "Test"},
        foreground=gui.Brush(),
        background=gui.Brush(),
        font=gui.Font(),
        selectable=True,
        status_tip="test",
        tool_tip="test",
        whats_this="test",
        checkstate="unchecked",
        size_hint=core.Size(10, 10),
        is_user_type=True,
    )


def test_standarditemmodel(qapp):
    model = gui.StandardItemModel()
    model.add("test")
    for _item in model:
        pass
    model = pickle_roundtrip(model)
    model += gui.StandardItem("Item")
    model[0]
    assert len(model.find_items("test")) == 1
    with pytest.raises(InvalidParamError):
        model.find_items("test", mode="wrong_mode")
    del model[0]
    model.add_item(
        "Test",
        icon="mdi.timer",
        data={1: "Test"},
        foreground=gui.Brush(),
        background=gui.Brush(),
        font=gui.Font(),
        selectable=True,
        status_tip="test",
        tool_tip="test",
        whats_this="test",
        checkstate="unchecked",
        size_hint=core.Size(10, 10),
        is_user_type=True,
    )


def test_statictext():
    text = gui.StaticText("test")
    repr(text)
    assert str(text) == "test"
    text.set_text_format("rich")
    with pytest.raises(InvalidParamError):
        text.set_text_format("test")
    assert text.get_text_format() == "rich"
    text.set_performance_hint("aggressive")
    with pytest.raises(InvalidParamError):
        text.set_performance_hint("test")
    assert text.get_performance_hint() == "aggressive"


def test_transform():
    transform = gui.Transform()
    for i, j in itertools.product(range(3), range(3)):
        assert transform[i, j] in [0, 1]
    bytes(transform)
    repr(transform)
    assert transform.get_type() == "none"
    transform = gui.Transform.clone_from(transform)


def test_textblock():
    block = gui.TextBlock()
    repr(block)
    assert 1 not in block


def test_textblockformat():
    block = gui.TextBlockFormat()
    block.set_alignment("left")
    with pytest.raises(InvalidParamError):
        block.set_alignment("test")
    assert block.get_alignment() == "left"
    block.set_marker("unchecked")
    with pytest.raises(InvalidParamError):
        block.set_marker("test")
    assert block.get_marker() == "unchecked"


def test_textblockgroup():
    doc = gui.TextDocument()
    group = gui.TextBlockGroup(doc)
    repr(group)
    for _textblock in group:
        pass


def test_textcursor():
    cursor = gui.TextCursor()
    cursor.set_position(1, "move")
    cursor.move_position("start", "move", 1)
    cursor.select("document")
    cursor.replace_text(0, 2, "test")
    cursor.select_text(1, 3)
    cursor.span()
    with cursor.edit_block():
        pass


def test_textdocument():
    doc = gui.TextDocument("This is a test\nHello")
    for i in doc:
        repr(i)
    assert doc[1].text() == "Hello"
    assert len(doc) == 2
    doc.set_text("test")
    doc.clear_stacks("undo_and_redo")
    with pytest.raises(InvalidParamError):
        doc.clear_stacks("test")
    # doc.add_resource("html")
    doc.set_default_cursor_move_style("logical")
    assert doc.get_default_cursor_move_style() == "logical"
    with pytest.raises(InvalidParamError):
        doc.set_default_cursor_move_style("test")
    doc.set_meta_information("document_title", "test")
    assert doc.get_meta_information("document_title") == "test"
    with pytest.raises(InvalidParamError):
        doc.set_meta_information("test", "test")
    with pytest.raises(InvalidParamError):
        doc.get_meta_information("test")


def test_paintdevice():
    device = gui.PaintDevice()
    device.get_metric("depth")
    with pytest.raises(InvalidParamError):
        device.get_metric("test")


def test_painter():
    class Test(widgets.Widget):
        def paintEvent(self, event):
            with gui.Painter(self) as painter:
                painter.use_antialiasing()
                painter.set_pen("none")
                painter.set_transparent_background(False)
                painter.set_transparent_background(True)
                painter.set_brush(gui.Brush())
                with painter.paint_on(widgets.Widget()):
                    pass
                painter.set_brush(gui.Color("red"))
                painter.fill_rect((0, 1, 3, 5), "transparent")
                painter.fill_rect(core.Rect(), "transparent")
                with pytest.raises(InvalidParamError):
                    painter.fill_rect(core.Rect(), "transparent", "test")
                with pytest.raises(ValueError):
                    painter.fill_rect(core.Rect(), "testus")
                painter.set_color("black")
                painter.set_composition_mode("source_atop")
                assert painter.get_composition_mode() == "source_atop"
                with pytest.raises(InvalidParamError):
                    painter.set_composition_mode("test")
                painter.set_clip_path(gui.PainterPath(), "replace")
                with pytest.raises(InvalidParamError):
                    painter.set_clip_path(gui.PainterPath(), "test")
                with pytest.raises(InvalidParamError):
                    painter.set_pen(style="test")
                with painter.backup_state():
                    pass
                assert painter.get_text_rect("test") is not None

    w = Test()
    w.repaint()
    # assert painter.get_composition_mode() == "source_atop"


def test_pagelayout():
    layout = gui.PageLayout()
    with pytest.raises(InvalidParamError):
        layout.set_orientation("test")
    layout.set_orientation("landscape")
    assert layout.get_orientation() == "landscape"
    with pytest.raises(InvalidParamError):
        layout.set_mode("test")
    layout.set_mode("full_page")
    assert layout.get_mode() == "full_page"
    with pytest.raises(InvalidParamError):
        layout.set_units("test")
    layout.set_units("pica")
    assert layout.get_units() == "pica"
    layout.get_page_size()


def test_pagesize():
    size = gui.PageSize()
    assert size.get_id() == "custom"
    with pytest.raises(ValueError):
        size.get_definition_units()
    size = gui.PageSize(gui.PageSize.PageSizeId.A3)
    size = pickle_roundtrip(size)
    assert size.get_definition_units() == "millimeter"


def test_painterpath():
    path = gui.PainterPath()
    rect = core.RectF(0, 0, 1, 1)
    path.addRect(rect)
    assert len(path) == 5
    assert bool(path)
    assert core.PointF(0.5, 0.5) in path
    path[1] = (0.5, 0.5)
    path.add_rect(core.QRect(0, 0, 1, 1))


def test_painterpathstroker():
    stroke = gui.PainterPathStroker()
    stroke.set_cap_style("round")
    with pytest.raises(InvalidParamError):
        stroke.set_cap_style("test")
    assert stroke.get_cap_style() == "round"
    stroke.set_join_style("bevel")
    with pytest.raises(InvalidParamError):
        stroke.set_join_style("test")
    assert stroke.get_join_style() == "bevel"


def test_palette():
    pal = gui.Palette()
    assert len(pal.get_colors()) == 21
    assert len(pal.get_brushes()) == 21
    color = gui.Color("red")
    pal.set_brush("window", "red")
    assert pal.get_brush("window") == color
    pal.highlight_inactive()
    pal.set_color("window", "red")
    pal["button"] = color
    assert pal["button"] == color
    bytes(pal)


def test_pdfwriter():
    writer = gui.PdfWriter("test")
    writer.set_desktop_resolution()
    writer.set_page_margins((0, 0, 0, 0), unit="pica")
    writer.set_pdf_version("v1_6")
    assert writer.get_pdf_version() == "v1_6"
    with pytest.raises(InvalidParamError):
        writer.set_pdf_version("test")


def test_pen():
    pen = gui.Pen()
    pen.set_color("blue")
    pen.set_cap_style("round")
    with pytest.raises(InvalidParamError):
        pen.set_cap_style("test")
    assert pen.get_cap_style() == "round"
    pen.set_join_style("bevel")
    with pytest.raises(InvalidParamError):
        pen.set_join_style("test")
    assert pen.get_join_style() == "bevel"
    pen.set_style("dash_dot")
    with pytest.raises(InvalidParamError):
        pen.set_style("test")
    assert pen.get_style() == "dash_dot"
    bytes(pen)


def test_picture():
    picture = gui.Picture()
    bytes(picture)


def test_pixmap(qapp):
    pix = gui.Pixmap()
    bytes(pix)
    pix.create_dot()
    pix.get_size()
    pix.get_rect()
    pix.to_image()


# def test_pixmapcache():
#     cache = gui.PixmapCache()
#     pix = gui.Pixmap()
#     cache["test"] = pix
#     cached = cache["test"]
#     assert pix.size() == cached.size()


def test_polygonf():
    poly = gui.PolygonF()
    poly.add_points((0, 0), (2, 0), (2, 1), (0, 1))
    poly2 = gui.PolygonF()
    poly2.add_points((1, 0), (3, 0), (3, 1), (1, 1))
    poly = pickle_roundtrip(poly)
    union = poly | poly2
    intersect = poly & poly2
    sub = union - intersect
    xor = poly ^ poly2
    assert sub == xor
    polygon = poly.to_polygon()
    assert type(polygon) is gui.Polygon
    poly.add_points((0, 1), core.PointF(2, 2))
    bytes(poly)


def test_polygon():
    rect1 = core.Rect(0, 0, 2, 1)
    rect2 = core.Rect(1, 0, 2, 1)
    poly = gui.Polygon(rect1, closed=True)
    poly2 = gui.Polygon(rect2, closed=True)
    intersect = poly & poly2
    expected = gui.Polygon(core.Rect(1, 0, 1, 1), closed=True)
    assert intersect == expected
    assert intersect.get_points() == [
        core.Point(1, 0),
        core.Point(2, 0),
        core.Point(2, 1),
        core.Point(1, 1),
        core.Point(1, 0),
    ]
    union = poly | poly2
    expected = gui.Polygon(core.Rect(0, 0, 3, 1), closed=True)
    assert list(union) == list(expected)
    sub = union - intersect
    xor = poly ^ poly2
    assert list(sub) == list(xor)
    # assert sub == xor
    poly = pickle_roundtrip(poly)
    poly.add_points((0, 1), core.Point(2, 2))
    assert bool(poly)
    assert core.Point(1, 0) in poly
    p = core.Point(5, 5)
    poly[5] = p
    assert poly[5] == p
    bytes(poly)


def test_region():
    region = gui.Region()
    bytes(region)


def test_regularexpressionvalidator():
    val = gui.RegularExpressionValidator()
    val.set_regex("[0-9]")
    val = pickle_roundtrip(val)
    assert val.get_regex() == "[0-9]"
    assert val.is_valid_value("0")


def test_syntaxhighlighter():
    doc = gui.TextDocument()
    gui.SyntaxHighlighter(doc)


def test_textcharformat():
    fmt = gui.TextCharFormat()
    fmt.set_font_weight("bold")
    assert fmt.get_font_weight() == "bold"
    fmt.set_foreground_color("yellow")
    fmt.set_background_color("yellow")
    with pytest.raises(InvalidParamError):
        fmt.set_font_weight("test")
    fmt = gui.TextCharFormat(bold=True)
    assert fmt.get_font_weight() == "bold"
    fmt.select_full_width()
    fmt.set_underline_style("dash")
    with pytest.raises(InvalidParamError):
        fmt.set_underline_style("test")
    assert fmt.get_underline_style() == "dash"
    fmt.set_font_style_hint("serif")
    with pytest.raises(InvalidParamError):
        fmt.set_font_style_hint("test")
    fmt.set_vertical_alignment("baseline")
    with pytest.raises(InvalidParamError):
        fmt.set_vertical_alignment("test")
    assert fmt.get_vertical_alignment() == "baseline"
    fmt.get_font()


def test_textimageformat():
    fmt = gui.TextImageFormat()
    assert bool(fmt)


def test_textlistformat():
    fmt = gui.TextListFormat()
    fmt.set_style("upper_roman")
    with pytest.raises(InvalidParamError):
        fmt.set_style("test")
    assert fmt.get_style() == "upper_roman"


def test_texttable():
    doc = gui.TextDocument()
    table = gui.TextTable(doc)
    with pytest.raises(IndexError):
        table[0, 0]
    with pytest.raises(IndexError):
        table[0]


def test_texttablecell():
    cell = gui.TextTableCell()
    assert cell
    # cell.get_format()


def test_texttableformat():
    fmt = gui.TextTableFormat()
    fmt.set_alignment("left")
    with pytest.raises(InvalidParamError):
        fmt.set_alignment("test")
    assert fmt.get_alignment() == "left"
    fmt.get_column_width_constraints()


def test_texttablecellformat():
    fmt = gui.TextTableCellFormat()
    fmt.set_bottom_border_style("dashed")
    with pytest.raises(InvalidParamError):
        fmt.set_bottom_border_style("test")
    assert fmt.get_bottom_border_style() == "dashed"

    fmt.set_left_border_style("dotted")
    with pytest.raises(InvalidParamError):
        fmt.set_left_border_style("test")
    assert fmt.get_left_border_style() == "dotted"

    fmt.set_right_border_style("solid")
    with pytest.raises(InvalidParamError):
        fmt.set_right_border_style("test")
    assert fmt.get_right_border_style() == "solid"

    fmt.set_top_border_style("double")
    with pytest.raises(InvalidParamError):
        fmt.set_top_border_style("test")
    assert fmt.get_top_border_style() == "double"

    fmt.set_border_style("groove")
    with pytest.raises(InvalidParamError):
        fmt.set_border_style("test")
    assert fmt.get_top_border_style() == "groove"

    fmt.get_bottom_border_brush()
    fmt.get_left_border_brush()
    fmt.get_right_border_brush()
    fmt.get_top_border_brush()


def test_textobject():
    doc = gui.TextDocument()
    obj = gui.TextObject(doc)
    repr(obj)
    obj.get_format()


def test_textlength():
    length = gui.TextLength()
    assert length.get_type() == "variable"
    repr(length)


def test_textframeformat():
    fmt = gui.TextFrameFormat()
    fmt.get_height()
    fmt.get_width()
    fmt.get_border_brush()
    fmt.set_border_style("dashed")
    with pytest.raises(InvalidParamError):
        fmt.set_border_style("test")
    assert fmt.get_border_style() == "dashed"
    fmt.set_page_break_policy("always_after")
    with pytest.raises(InvalidParamError):
        fmt.set_page_break_policy("test")
    assert fmt.get_page_break_policy() == "always_after"
    fmt.set_position("flow_left")
    with pytest.raises(InvalidParamError):
        fmt.set_position("test")
    assert fmt.get_position() == "flow_left"


def test_textformat():
    fmt = gui.TextFormat()
    fmt[1] = "test"
    assert fmt[1] == "test"
    assert 1 in fmt
    assert not bool(fmt)
    repr(fmt)
    fmt.get_background()
    fmt.get_foreground()
    fmt.get_brush_property(1)
    fmt.get_pen_property(1)
    fmt.get_color_property(1)
    fmt.set_layout_direction("right_to_left")
    with pytest.raises(InvalidParamError):
        fmt.set_layout_direction("test")
    assert fmt.get_layout_direction() == "right_to_left"


def test_textframe():
    doc = gui.TextDocument()
    frame = gui.TextFrame(doc)
    repr(frame)
    frame.get_first_cursor_position()
    frame.get_last_cursor_position()


def test_undocommand():
    cmd = gui.UndoCommand()
    cmd2 = gui.UndoCommand(cmd)
    assert cmd[0] == cmd2
    assert len(cmd) == 1


def test_undogroup():
    group = gui.UndoGroup()
    stack = gui.UndoStack()
    group.addStack(stack)
    assert len(group) == 1
    assert group[0] == stack


def test_undostack():
    stack = gui.UndoStack()
    cmd = stack.add_command("test", redo=lambda: print("a"), undo=lambda: print("b"))
    assert stack[0] == cmd
    assert len(stack) == 1
    with stack.create_macro("test"):
        pass


def test_window():
    wnd = gui.Window()
    assert wnd.get_surface_class() == "window"
    assert wnd.get_surface_type() == "raster"
    wnd.set_visibility("maximized")
    with pytest.raises(InvalidParamError):
        wnd.set_visibility("test")
    assert wnd.get_visibility() == "maximized"
    wnd.start_system_resize("right")
    with pytest.raises(InvalidParamError):
        wnd.start_system_resize("test")


def test_validator():
    gui.Validator()


def test_vector3d():
    vector = gui.Vector3D(0, 0, 1)
    assert abs(vector) == 1.0
    assert bool(vector)
    vector.to_point()
    vector.to_pointf()


def test_vector4d():
    vector = gui.Vector4D(0, 0, 0, 1)
    assert abs(vector) == 1.0
    assert bool(vector)
    vector.to_point()
    vector.to_pointf()



from __future__ import annotations

from typing import TYPE_CHECKING, ClassVar

from prettyqt import constants, gui, itemmodels


if TYPE_CHECKING:
    from collections.abc import Sequence


class WhatsThisColumn(itemmodels.ColumnItem):
    name = "Whats this"
    doc = "Whats This"
    editable = True

    def get_data(self, item: gui.QShortcut, role: constants.ItemDataRole):
        match role:
            case constants.DISPLAY_ROLE | constants.EDIT_ROLE:
                return item.whatsThis()

    def set_data(self, item, value, role):
        match role:
            case constants.EDIT_ROLE:
                return item.setWhatsThis(value)


class EnabledColumn(itemmodels.ColumnItem):
    name = "Enabled"
    doc = "Whether shortcut is enabled."
    checkable = True

    def get_data(self, item: gui.QShortcut, role: constants.ItemDataRole):
        match role:
            case constants.CHECKSTATE_ROLE:
                return self.to_checkstate(item.isEnabled())

    def set_data(self, item, value, role):
        match role:
            case constants.CHECKSTATE_ROLE:
                item.setEnabled(not item.isEnabled())
                return True
        return False


class ShortcutColumn(itemmodels.ColumnItem):
    name = "Shortcut"
    doc = "Keyboard shortcut."
    editable = True

    def get_data(self, item: gui.QShortcut, role: constants.ItemDataRole):
        match role:
            case constants.DISPLAY_ROLE:
                return item.key().toString()
            case constants.EDIT_ROLE:
                return item.key()

    def set_data(self, item, value, role):
        match role:
            case constants.EDIT_ROLE:
                item.setKey(value)
                return True
        return False


class ContextColumn(itemmodels.ColumnItem):
    name = "Context"
    doc = "Shortcut context."
    editable = True

    def get_data(self, item: gui.QShortcut, role: constants.ItemDataRole):
        match role:
            case constants.DISPLAY_ROLE:
                return constants.SHORTCUT_CONTEXT.inverse[item.context()]
            case constants.EDIT_ROLE:
                return item.context()

    def set_data(self, item, value, role):
        match role:
            case constants.EDIT_ROLE:
                item.setContext(value)
                return True
        return False


class AutoRepeatColumn(itemmodels.ColumnItem):
    name = "Auto-repeat"
    doc = "Auto-repeat."
    checkable = True

    def get_data(self, item: gui.QShortcut, role: constants.ItemDataRole):
        match role:
            case constants.CHECKSTATE_ROLE:
                return self.to_checkstate(item.autoRepeat())

    def set_data(self, item, value, role):
        match role:
            case constants.CHECKSTATE_ROLE:
                item.setAutoRepeat(not item.autoRepeat())


class ParentColumn(itemmodels.ColumnItem):
    name = "Parent widget"
    doc = "Parent widget."

    def get_data(self, item: gui.QShortcut, role: constants.ItemDataRole):
        match role:
            case constants.DISPLAY_ROLE:
                return item.parent().windowTitle() or item.parent().__class__.__name__


class ShortcutsModel(itemmodels.ColumnTableModel):
    """Model to display a list of gui.QShortcuts."""

    COLUMNS: ClassVar = [
        WhatsThisColumn,
        EnabledColumn,
        ShortcutColumn,
        AutoRepeatColumn,
        ContextColumn,
        ParentColumn,
    ]
    SUPPORTS = gui.QShortcut

    def __init__(self, shortcuts: Sequence[gui.QShortcut], parent=None):
        super().__init__(shortcuts, self.COLUMNS, parent=parent)

    @classmethod
    def supports(cls, instance) -> bool:
        match instance:
            case (gui.QShortcut(), *_):
                return True
            case _:
                return False


if __name__ == "__main__":
    from prettyqt import widgets

    app = widgets.app()
    view = widgets.TreeView()
    shortcuts = [
        gui.Shortcut("Ctrl+A", view, print, object_name="super duper"),
        gui.Shortcut("Ctrl+B", view, print, object_name="this is an shortcut"),
        gui.Shortcut("Ctrl+C", view, print, object_name="fsfsfsdfs sfd"),
        gui.Shortcut("Ctrl+D", view, print, object_name="fsfsfsdfs sfd"),
    ]
    for i in shortcuts:
        i.setWhatsThis("abc")
    model = ShortcutsModel(shortcuts, parent=view)
    view.setModel(model)
    view.set_delegate("editor")
    view.resize(640, 480)
    view.setEditTriggers(view.EditTrigger.AllEditTriggers)
    view.set_selection_behavior("rows")
    view.adapt_sizes()
    view.show()
    with app.debug_mode():
        app.exec()



from collections.abc import Callable, Iterable
from typing import Any

from prettyqt import constants, core


class ListMixin:
    SORT_METHODS: dict[int, Callable]
    MIME_TYPE: str = ""

    def __init__(self, *args, items=None, **kwargs):
        super().__init__(*args, **kwargs)  # type: ignore
        self.items = items or []

    def setData(
        self,
        index: core.ModelIndex,
        value: Any,
        role: constants.ItemDataRole = constants.EDIT_ROLE,
    ) -> bool:
        if role == constants.USER_ROLE:
            self.items[index.row()] = value
            self.update_row(index.row())
            return True
        return super().setData(index, value, role)  # type: ignore

    def removeRows(self, row: int, count: int, parent):
        # called by default implementation of QAbstractItemModel::startDrag
        end_row = row + count - 1
        with self.remove_rows(row, end_row, parent):
            for i in range(end_row, row - 1, -1):
                self.items.pop(i)
        return True

    # def insertRows(self, row: int, count: int, parent):
    #     # called by default implementation of QAbstractItemModel::dropMimeData
    #     end_row = row + count - 1
    #     with self.insert_rows(row, end_row, parent):
    #         for i in range(end_row, row - 1, -1):
    #             self.items.insert(i,)
    #     return True

    def rowCount(self, parent: core.ModelIndex | None = None) -> int:
        """Required override for AbstractitemModels."""
        parent = parent or core.ModelIndex()
        return 0 if parent.column() > 0 or parent.isValid() else len(self.items)

    def data_by_index(self, index):
        return self.items[index.row()]

    def dropMimeData(self, mime_data, action, row, column, parent_index):
        if not mime_data.hasFormat(self.MIME_TYPE):
            return False
        # Since we only drop in between items, parent_index must be invalid,
        # and we use the row arg to know where the drop took place.
        if parent_index.isValid():
            return False
        indexes = mime_data.get_json_data(self.MIME_TYPE)
        pos = row if row < len(self.items) and row != -1 else len(self.items)
        rem_offset = sum(i <= pos for i in indexes)
        new = [self.items[i] for i in indexes]
        with self.change_layout():
            for i in sorted(indexes, reverse=True):
                self.items.pop(i)
            for item in reversed(new):
                self.items.insert(pos - rem_offset, item)
        return False

    def sort(self, ncol: int, order):
        """Sort table by given column number."""
        is_asc = order == constants.ASCENDING
        if sorter := self.SORT_METHODS.get(ncol):
            with self.change_layout():
                self.items.sort(key=sorter, reverse=is_asc)

    def add(self, item: Any, position: int | None = None):
        """Append provided item to the list."""
        self.add_items(items=[item], position=position)
        return item

    def add_items(self, items: Iterable[Any], position: int | None = None):
        """Append a list of items to the list."""
        if position is None:
            position = len(self.items)
        items = list(items)
        with self.insert_rows(position, position + len(items) - 1):
            for i in range(len(items)):
                self.items.insert(i + position, items[i])
            # self.items.extend(items)
        return items

    def remove_items(self, offsets: Iterable[int]):
        for offset in sorted(offsets, reverse=True):
            self.removeRow(offset)

    def supportedDropActions(self):
        return constants.MOVE_ACTION

    def mimeTypes(self):
        return [self.MIME_TYPE]

    def mimeData(self, indexes):
        """AbstractItemModel override, defines the data used for drag and drop.

        atm this just returns the positions (not sure if this is perfect)
        """
        mime_data = core.MimeData()
        data = [i.row() for i in indexes if i.column() == 0]
        mime_data.set_json_data(self.MIME_TYPE, data)
        return mime_data

    # list interface

    def pop(self, row=None):
        if row is None:
            row = len(self.items) - 1
        result = self.items[row]
        self.removeRow(row)
        return result

    def __getitem__(self, row):
        return self.items[row]

    def __setitem__(self, row: int | slice, value):
        match row:
            case slice():
                rng = range(row.start or 0, row.stop or len(self.items), row.step or 1)
                for count, i in enumerate(rng):
                    if i < self.rowCount():
                        index = self.index(i)
                        self.setData(i, value[count], role=constants.USER_ROLE)
                    else:
                        self.items.append(value[count])
            case int():
                index = self.index(row)
                self.setData(index, value, role=constants.USER_ROLE)
            case _:
                raise ValueError(row)

    def __len__(self):
        return len(self.items)

    def insert(self, row: int, value):
        with self.insert_row(row):
            self.items.insert(row, value)

    def append(self, value):
        row = len(self.items)
        self.insert(row, value)

    def extend(self, values):
        pos = len(self.items)
        with self.insert_rows(pos, pos + len(values)):
            self.items.extend(values)

    def set_list(self, values):
        """Set the model to a new list."""
        with self.reset_model():
            self.items = values

    def remove(self, item):
        if item in self.items:
            pos = self.items.index(item)
            with self.remove_row(pos):
                self.items.remove(item)


if __name__ == "__main__":
    from prettyqt import debugging, widgets

    class Test(ListMixin, core.AbstractTableModel):
        def data(self, index, role):
            if role == constants.DISPLAY_ROLE:
                return self.items[index.row()]
            return None

        def columnCount(self, index: core.ModelIndex | None = None):
            return 1

    model = Test()
    model.set_list(["a", "b"])
    app = widgets.app()
    view = widgets.TableView()
    stalker = debugging.Stalker(model)
    view.set_model(model)
    view.show()
    app.sleep(2)
    model.pop(0)
    with app.debug_mode():
        app.exec()



from __future__ import annotations

from prettyqt.qt import QtDesigner


class AbstractExtensionFactory(QtDesigner.QAbstractExtensionFactory):
    """Interface for extension factories in Qt Designer."""


if __name__ == "__main__":
    fact = AbstractExtensionFactory()



from __future__ import annotations

from prettyqt import gui


class HexValidator(gui.Validator):
    """Validator which checks for hexadecimal values."""

    ID = "hex"

    def __init__(self, maximum: int | None = None, parent=None):
        super().__init__(parent)
        self._maximum = maximum

    def __eq__(self, other: object):
        return isinstance(other, HexValidator) and other._maximum == self._maximum

    def validate(self, text: str, pos: int = 0) -> tuple[gui.QValidator.State, str, int]:
        try:
            val = int(text, 0)
        except ValueError:
            return self.State.Intermediate, text, pos
        if self._maximum is not None and val > self._maximum:
            return self.State.Invalid, text, pos
        return self.State.Acceptable, text, pos


if __name__ == "__main__":
    from prettyqt import widgets

    val = HexValidator()
    app = widgets.app()
    widget = widgets.LineEdit()
    widget.setValidator(val)
    widget.show()
    with app.debug_mode():
        app.exec()
    print(widget.hasAcceptableInput())



from __future__ import annotations

from collections.abc import MutableMapping
from typing import Literal

from prettyqt.qt import QtMultimedia
from prettyqt.utils import bidict, datatypes, get_repr


KeyStr = Literal[
    "title",
    "author",
    "comment",
    "description",
    "genre",
    "date",
    "language",
    "publisher",
    "copyright",
    "url",
    "duration",
    "media_type",
    "file_format",
    "audio_bit_rate",
    "audio_codec",
    "video_bitrate",
    "video_codec",
    "video_frame_rate",
    "album_title",
    "album_artist",
    "contributing_artist",
    "track_number",
    "composer",
    "lead_performer",
    "thumbnail_image",
    "cover_art_image",
    "orientation",
    "resolution",
]

KEY: bidict[KeyStr, QtMultimedia.QMediaMetaData.Key] = bidict(
    title=QtMultimedia.QMediaMetaData.Key.Title,
    author=QtMultimedia.QMediaMetaData.Key.Author,
    comment=QtMultimedia.QMediaMetaData.Key.Comment,
    description=QtMultimedia.QMediaMetaData.Key.Description,
    genre=QtMultimedia.QMediaMetaData.Key.Genre,
    date=QtMultimedia.QMediaMetaData.Key.Date,
    language=QtMultimedia.QMediaMetaData.Key.Language,
    publisher=QtMultimedia.QMediaMetaData.Key.Publisher,
    copyright=QtMultimedia.QMediaMetaData.Key.Copyright,
    url=QtMultimedia.QMediaMetaData.Key.Url,
    duration=QtMultimedia.QMediaMetaData.Key.Duration,
    media_type=QtMultimedia.QMediaMetaData.Key.MediaType,
    file_format=QtMultimedia.QMediaMetaData.Key.FileFormat,
    audio_bit_rate=QtMultimedia.QMediaMetaData.Key.AudioBitRate,
    audio_codec=QtMultimedia.QMediaMetaData.Key.AudioCodec,
    video_bitrate=QtMultimedia.QMediaMetaData.Key.VideoBitRate,
    video_codec=QtMultimedia.QMediaMetaData.Key.VideoCodec,
    video_frame_rate=QtMultimedia.QMediaMetaData.Key.VideoFrameRate,
    album_title=QtMultimedia.QMediaMetaData.Key.AlbumTitle,
    album_artist=QtMultimedia.QMediaMetaData.Key.AlbumArtist,
    contributing_artist=QtMultimedia.QMediaMetaData.Key.ContributingArtist,
    track_number=QtMultimedia.QMediaMetaData.Key.TrackNumber,
    composer=QtMultimedia.QMediaMetaData.Key.Composer,
    lead_performer=QtMultimedia.QMediaMetaData.Key.LeadPerformer,
    thumbnail_image=QtMultimedia.QMediaMetaData.Key.ThumbnailImage,
    cover_art_image=QtMultimedia.QMediaMetaData.Key.CoverArtImage,
    orientation=QtMultimedia.QMediaMetaData.Key.Orientation,
    resolution=QtMultimedia.QMediaMetaData.Key.Resolution,
)


class MediaMetaData(
    QtMultimedia.QMediaMetaData,
    MutableMapping,
    metaclass=datatypes.QABCMeta,
):
    def __repr__(self):
        return get_repr(self, dict(self))

    def __setitem__(self, key: KeyStr, value: datatypes.Variant):
        if isinstance(key, str):
            key = KEY[key]
        self.insert(key, value)

    def __getitem__(self, key: KeyStr) -> datatypes.Variant:
        if isinstance(key, str):
            key = KEY[key]
        if key not in self.keys():
            raise KeyError(key)
        return self.value(key)

    def __delitem__(self, key: KeyStr):
        if isinstance(key, str):
            key = KEY[key]
        self.remove(key)

    def __bool__(self):
        return not self.isEmpty()

    def __iter__(self):
        return iter(self.keys())


if __name__ == "__main__":
    from prettyqt import core

    metadata = MediaMetaData()
    metadata["resolution"] = core.QSize(10, 10)
    print(metadata)



from __future__ import annotations

import logging
from typing import TYPE_CHECKING, Self

from prettyqt.utils import baseresolver, get_repr


if TYPE_CHECKING:
    from collections.abc import Callable, Iterable, Iterator, Sequence


logger = logging.getLogger(__name__)


class NodeResolver(baseresolver.BaseResolver):
    def __init__(self, path_attr: str = "obj", ignore_case: bool = False):
        """Resolve any `Node` paths using attribute `path_attr`.

        Arguments:
            path_attr: Name of the node attribute to be used for resolving
            ignore_case: Enable case insensisitve handling.
        """
        super().__init__(ignore_case=ignore_case)
        self.path_attr = path_attr

    def get_parent(self, node):
        return node.parent

    def get_children(self, node):
        return node.children

    def get_root(self, node):
        return node.root

    def get_attribute(self, node):
        return getattr(node, self.path_attr)


class BaseNode:
    """A Node class that can be used to build trees."""

    # __slots__ = ("parent_item", "children")

    def __init__(self, parent: Self | None = None):
        self.parent_item = parent
        self.children: list[Self] = []

    def __repr__(self):
        return get_repr(self)

    def __iter__(self) -> Iterator[Self]:
        return iter(self.children)

    def __getitem__(self, index: int) -> Self:
        return self.children[index]

    def __rshift__(self, other: Self):
        """Set children using >> bitshift operator for self >> other.

        Args:
            other (Self): other node, children
        """
        other.parent_item = self

    def __lshift__(self, other: Self):
        """Set parent using << bitshift operator for self << other.

        Args:
            other (Self): other node, parent
        """
        self.parent_item = other

    def __copy__(self) -> Self:
        """Shallow copy self."""
        obj = type(self).__new__(self.__class__)
        obj.__dict__.update(self.__dict__)
        return obj

    def append_child(self, item: Self):
        item.parent_item = self
        self.children.append(item)

    def insert_children(self, idx: int, items: Sequence[Self]):
        self.children[idx:idx] = items
        for item in items:
            item.parent_item = self

    def parent(self) -> Self | None:
        return self.parent_item

    @property
    def ancestors(self) -> Iterable[Self]:
        """Get iterator to yield all ancestors of self, does not include self."""
        node = self
        while (node := node.parent_item) is not None:
            yield node

    @property
    def descendants(self) -> Iterable[Self]:
        """Get iterator to yield all descendants of self, does not include self."""
        yield from preorder_iter(self, filter_condition=lambda _node: _node != self)

    @property
    def leaves(self) -> Iterable[Self]:
        """Get iterator to yield all leaf nodes from self."""
        yield from preorder_iter(self, filter_condition=lambda _node: _node.is_leaf)

    @property
    def siblings(self) -> Iterable[Self]:
        """Get siblings of self."""
        if self.parent_item is None:
            return ()
        return tuple(child for child in self.parent_item.children if child is not self)

    @property
    def left_sibling(self) -> Self | None:
        """Get sibling left of self."""
        if self.parent_item:
            children = self.parent_item.children
            if child_idx := children.index(self):
                return self.parent_item.children[child_idx - 1]
            return None
        return None

    @property
    def right_sibling(self) -> Self | None:
        """Get sibling right of self."""
        if self.parent_item:
            children = self.parent_item.children
            child_idx = children.index(self)
            if child_idx + 1 < len(children):
                return self.parent_item.children[child_idx + 1]
            return None
        return None

    @property
    def node_path(self) -> Iterable[Self]:
        """Get tuple of nodes starting from root."""
        if self.parent_item is None:
            return [self]
        return (*list(self.parent_item.node_path), self)

    @property
    def is_root(self) -> bool:
        """Get indicator if self is root node."""
        return self.parent_item is None

    @property
    def is_leaf(self) -> bool:
        """Get indicator if self is leaf node."""
        return not len(list(self.children))

    @property
    def root(self) -> Self:
        """Get root node of tree."""
        return self if self.parent_item is None else self.parent_item.root

    @property
    def depth(self) -> int:
        """Get depth of self, indexing starts from 1."""
        return 1 if self.parent_item is None else self.parent_item.depth + 1

    @property
    def max_depth(self) -> int:
        """Get maximum depth from root to leaf node."""
        return max(
            [self.root.depth] + [node.depth for node in list(self.root.descendants)]
        )

    def row(self) -> int:
        return self.parent_item.children.index(self) if self.parent_item else 0

    def pretty_print(self, indent: int = 0):
        text = indent * "    " + repr(self)
        logger.info(text)
        for child_item in self.children:
            child_item.pretty_print(indent + 1)


class Node(BaseNode):
    """Like BaseNode, but with one object attached to the node."""

    # __slots__ = ("obj",)

    def __init__(self, obj, parent: Self | None = None):
        super().__init__(parent)
        self.obj = obj

    def __repr__(self):
        return get_repr(self, self.obj)


def preorder_iter(
    tree: BaseNode,
    filter_condition: Callable[[BaseNode], bool] | None = None,
    stop_condition: Callable[[BaseNode], bool] | None = None,
    max_depth: int = 0,
) -> Iterable[BaseNode]:
    """Iterate through all children of a tree.

    Pre-Order Iteration Algorithm, NLR
        1. Visit the current node.
        2. Recursively traverse the current node's left subtree.
        3. Recursively traverse the current node's right subtree.

    It is topologically sorted because a parent node is processed before its child nodes.

    >>> path_list = ["a/b/d", "a/b/e/g", "a/b/e/h", "a/c/f"]
    >>> root = list_to_tree(path_list)
    >>> print_tree(root)
    a
    ├── b
    │   ├── d
    │   └── e
    │       ├── g
    │       └── h
    └── c
        └── f

    >>> [node.node_name for node in preorder_iter(root)]
    ['a', 'b', 'd', 'e', 'g', 'h', 'c', 'f']

    >>> [node.node_name for node in preorder_iter(root,
    filter_condition=lambda x: x.node_name in ["a", "d", "e", "f", "g"])]
    ['a', 'd', 'e', 'g', 'f']

    >>> [node.node_name for node in preorder_iter(root,
    stop_condition=lambda x: x.node_name=="e")]
    ['a', 'b', 'd', 'c', 'f']

    >>> [node.node_name for node in preorder_iter(root, max_depth=3)]
    ['a', 'b', 'd', 'e', 'c', 'f']

    Args:
        tree: input tree
        filter_condition: function that takes in node as argument, optional
            Return node if condition evaluates to `True`
        stop_condition: function that takes in node as argument, optional
            Stops iteration if condition evaluates to `True`
        max_depth: maximum depth of iteration, based on `depth` attribute, optional
    """
    if (not max_depth or tree.depth <= max_depth) and (
        not stop_condition or not stop_condition(tree)
    ):
        if not filter_condition or filter_condition(tree):
            yield tree
        for child in tree.children:
            yield from preorder_iter(child, filter_condition, stop_condition, max_depth)


if __name__ == "__main__":
    model = Node("test")
    model.parent()



from __future__ import annotations

import os
from typing import TYPE_CHECKING

from prettyqt import core, gui


if TYPE_CHECKING:
    from prettyqt.utils import datatypes


class DesktopServices(gui.QDesktopServices):
    """Methods for accessing common desktop services."""

    @classmethod
    def open_url(cls, location: datatypes.PathType | datatypes.UrlType) -> bool:
        if not isinstance(location, core.QUrl):
            location = core.Url.from_user_input(os.fspath(location))
        return cls.openUrl(location)



from __future__ import annotations

import logging
from typing import Any

from prettyqt import constants, core


logger = logging.getLogger(__name__)


class StackProxyModel(core.AbstractProxyModel):
    """Proxy model to stack columns into rows.

    Transforms wide format with multiple columns into long format
    by stacking columns vertically. Similar to pandas.DataFrame.stack().

    === "Without proxy"

        ```py
        app = widgets.app()
        data = {
            ("Sales", "2020"): [100, 200],
            ("Sales", "2021"): [150, 250],
            ("Costs", "2020"): [80, 160],
            ("Costs", "2021"): [90, 180],
        }
        model = gui.StandardItemModel.from_dict(data)
        table = widgets.TableView()
        table.set_model(model)
        table.show()
        ```

    === "With proxy"

        ```py
        app = widgets.app()
        data = {
            ("Sales", "2020"): [100, 200],
            ("Sales", "2021"): [150, 250],
            ("Costs", "2020"): [80, 160],
            ("Costs", "2021"): [90, 180],
        }
        model = gui.StandardItemModel.from_dict(data)
        table = widgets.TableView()
        table.set_model(model)
        table.proxifier.stack()
        table.show()
        ```
    """

    ID = "stack"
    ICON = "mdi6.table-pivot"

    def __init__(self, level: int | None = None, **kwargs: Any):
        """Initialize StackProxyModel.

        Args:
            level: If headers are multi-level, which level to stack by.
                  None means stack all columns.
            kwargs: Additional keyword arguments.
        """
        super().__init__(**kwargs)
        self._level = level

    def rowCount(self, parent: core.ModelIndex | None = None) -> int:
        """Return number of rows in stacked format."""
        source = self.sourceModel()
        if not source:
            return 0
        return source.rowCount() * source.columnCount()

    def columnCount(self, parent: core.ModelIndex | None = None) -> int:
        """Return number of columns (always 2 - index and values)."""
        return 2

    def data(
        self,
        index: core.ModelIndex,
        role: constants.ItemDataRole = constants.DISPLAY_ROLE,
    ) -> Any:
        """Return data for given index and role."""
        if not index.isValid():
            return None

        source = self.sourceModel()
        total_cols = source.columnCount()

        # Calculate original row and column
        source_row = index.row() // total_cols
        source_col = index.row() % total_cols

        # First column shows the header name
        if index.column() == 0:
            return source.headerData(source_col, constants.HORIZONTAL, role)

        # Second column shows the value
        source_index = source.index(source_row, source_col)
        return source.data(source_index, role)

    def headerData(
        self,
        section: int,
        orientation: constants.Orientation,
        role: constants.ItemDataRole = constants.DISPLAY_ROLE,
    ) -> Any:
        """Return header data."""
        if orientation != constants.HORIZONTAL or role != constants.DISPLAY_ROLE:
            return None

        return ["Variable", "Value"][section]

    def mapToSource(self, proxy_index: core.ModelIndex) -> core.ModelIndex:
        """Map proxy index to source index."""
        if not proxy_index.isValid() or proxy_index.column() != 1:
            return core.ModelIndex()

        source = self.sourceModel()
        total_cols = source.columnCount()

        source_row = proxy_index.row() // total_cols
        source_col = proxy_index.row() % total_cols

        return source.index(source_row, source_col)

    def mapFromSource(self, source_index: core.ModelIndex) -> core.ModelIndex:
        """Map source index to proxy index."""
        if not source_index.isValid():
            return core.ModelIndex()

        proxy_row = (
            source_index.row() * self.sourceModel().columnCount() + source_index.column()
        )
        return self.index(proxy_row, 1)


if __name__ == "__main__":
    from prettyqt import gui, widgets

    app = widgets.app(style="Windows")
    data = {
        ("Sales", "2020"): [100, 200],
        ("Sales", "2021"): [150, 250],
        ("Costs", "2020"): [80, 160],
        ("Costs", "2021"): [90, 180],
    }
    model = gui.StandardItemModel.from_dict(data)
    table = widgets.TableView()
    table.set_model(model)
    table.setWindowTitle("StackProxyModel example")
    table.set_icon("mdi6.table-pivot")
    table.show()
    table.resize(600, 400)
    table.h_header.resize_sections("stretch")

    # Stack columns into rows
    table.proxifier.stack()

    with app.debug_mode():
        app.exec()



from __future__ import annotations

import logging

from prettyqt import core, widgets
from prettyqt.utils import colors, datatypes, helpers


logger = logging.getLogger(__name__)


class AnimationTimer(core.Timer):
    """Timer with animation attached.

    Is returned for animations in order to stop them afterwards.
    """

    def __init__(
        self,
        animation: core.PropertyAnimation,
        interval: int,
        single_shot: bool = True,
        parent: core.QObject | None = None,
    ):
        # we could also take targetObject as parent?
        # that would limit the usage to PropertyAnimations though.
        self.animation = animation
        super().__init__(
            timeout=animation.start,
            interval=interval,
            single_shot=single_shot,
            parent=parent,
        )


class AnimationWrapper:
    """Wrapper for property animations."""

    def __init__(self, prop_name: str, fx: Fx):
        self.fx = fx
        self._widget = fx._widget
        self._property_name = prop_name

    def animate(
        self,
        start: datatypes.VariantType,
        end: datatypes.VariantType,
        easing: core.easingcurve.TypeStr = "in_out_sine",
        duration: int = 1000,
        delay: int | str = 0,
        reverse: bool = False,
        single_shot: bool = True,
    ) -> AnimationTimer:
        """General animation with global coordinates."""
        self._animation = core.PropertyAnimation()
        prop = getattr(self._widget, self._property_name)
        self._animation.apply_to(prop)
        self._animation.set_start_value(start)
        self._animation.set_end_value(end)
        self._animation.set_easing(easing)
        self._animation.set_duration(duration)
        logger.debug(f"Setting up animation: {start=}, {end=}, {easing=}, {duration=}")
        if reverse:
            self._animation.append_reversed()
        return self.fx.run(self._animation, delay=delay, single_shot=single_shot)

    def transition_to(
        self,
        end: datatypes.VariantType | tuple,
        easing: core.easingcurve.TypeStr = "in_out_sine",
        duration: int = 1000,
        delay: int | str = 0,
        relative: bool = False,
        reverse: bool = False,
        single_shot: bool = True,
    ) -> AnimationTimer:
        """Makes property transition from current value to given end value."""
        self._animation = core.PropertyAnimation()
        prop = getattr(self._widget, self._property_name)
        self._animation.apply_to(prop)
        prop_name = self._animation.get_property_name()
        obj = self._animation.targetObject()
        start: datatypes.VariantType = obj.property(prop_name)
        end = datatypes.align_types(start, end)
        if isinstance(end, core.QPoint | core.QRect | core.QPointF | core.QRectF):
            end = obj.map_to("parent", end)
        if relative and hasattr(end, "__add__"):
            end = end + start
        return self.animate(
            start,
            end,
            easing=easing,
            duration=duration,
            delay=delay,
            reverse=reverse,
            single_shot=single_shot,
        )

    def transition_from(
        self,
        start: datatypes.VariantType,
        easing: core.easingcurve.TypeStr = "in_out_sine",
        duration: int = 1000,
        delay: int | str = 0,
        relative: bool = False,
        reverse: bool = False,
        single_shot: bool = True,
    ) -> AnimationTimer:
        """Makes property transition from given start value to its current value."""
        self._animation = core.PropertyAnimation()
        prop = getattr(self._widget, self._property_name)
        self._animation.apply_to(prop)
        prop_name = self._animation.get_property_name()
        obj = self._animation.targetObject()
        end: core.VariantType = obj.property(prop_name)
        start = datatypes.align_types(end, start)
        if isinstance(start, core.QPoint | core.QRect | core.QPointF | core.QRectF):
            start = obj.map_to("parent", start)
        if relative and hasattr(start, "__add__"):
            start = end + start
        return self.animate(
            start,
            end,
            easing=easing,
            duration=duration,
            delay=delay,
            reverse=reverse,
            single_shot=single_shot,
        )

    def animate_on_event(
        self,
        event: core.event.TypeStr,
        start: datatypes.VariantType | None,
        end: datatypes.VariantType,
        easing: core.easingcurve.TypeStr = "in_out_sine",
        duration: int = 1000,
        delay: int = 0,
        reverse: bool = False,
    ) -> core.PropertyAnimation:
        """Starts property transition from start to end when given event occurs."""
        self._animation = core.PropertyAnimation()
        prop = getattr(self._widget, self._property_name)
        self._animation.apply_to(prop)
        if start is None:
            prop_name = self._animation.get_property_name()
            obj = self._animation.targetObject()
            start = obj.property(prop_name)
            self._animation.setStartValue(start)
            end = datatypes.align_types(start, end)
            if isinstance(end, core.QPoint | core.QRect | core.QPointF | core.QRectF):
                end = obj.map_to("parent", end)
        else:
            self._animation.setStartValue(start)
        self._animation.setEndValue(end)
        self._animation.setDuration(duration)
        self._animation.set_easing(easing)
        if reverse:
            self._animation.append_reversed()

        def on_event(event):
            self.fx.run(self._animation, delay=delay)
            return False

        obj = self._animation.targetObject()
        obj.add_callback_for_event(on_event, include=event)
        return self._animation


class Fx:
    """Fx delegator which allows a quick and easy way to animate widgets."""

    def __init__(self, widget: widgets.QWidget):
        self._widget = widget
        self._wrapper = None
        # self._meta = core.MetaObject(self._widget.metaObject())

    def __getitem__(self, value: str) -> AnimationWrapper:
        value = helpers.to_lower_camel(value)
        logger.debug(f"Building {value!r} PropertyAnimation for {self._widget!r}")
        wrapper = AnimationWrapper(value, self)
        self._wrapper = wrapper
        return wrapper

    # def __getattr__(self, attr):
    #     return self.__getitem__(attr)

    def set_colorize_effect(
        self,
        color: datatypes.ColorType,
        strength: float = 0.7,
    ) -> widgets.GraphicsColorizeEffect:
        effect = widgets.GraphicsColorizeEffect(self._widget)
        effect.setColor(colors.get_color(color))
        effect.setStrength(strength)
        self._widget.setGraphicsEffect(effect)
        return effect

    def set_opacity_effect(self, opacity: float) -> widgets.GraphicsOpacityEffect:
        effect = widgets.GraphicsOpacityEffect(self._widget)
        effect.setOpacity(opacity)
        self._widget.setGraphicsEffect(effect)
        return effect

    def set_blur_effect(self, radius: int) -> widgets.GraphicsBlurEffect:
        effect = widgets.GraphicsBlurEffect(self._widget)
        effect.setBlurRadius(radius)
        self._widget.setGraphicsEffect(effect)
        return effect

    def set_drop_shadow_effect(
        self,
        radius: int = 10,
        color: datatypes.ColorType = "blue",
    ) -> widgets.GraphicsDropShadowEffect:
        effect = widgets.GraphicsDropShadowEffect(self._widget)
        effect.setBlurRadius(radius)
        effect.setColor(colors.get_color(color))
        self._widget.setGraphicsEffect(effect)
        return effect

    def fade_in(
        self,
        duration: int = 1000,
        easing: core.easingcurve.TypeStr = "in_out_sine",
        delay: int = 0,
    ) -> AnimationTimer:
        """Trigger a fade-in animation."""
        return self["windowOpacity"].transition_from(
            0.0,
            easing=easing,
            duration=duration,
            delay=delay,
        )

    def fade_out(
        self,
        duration: int = 1000,
        easing: core.easingcurve.TypeStr = "in_out_sine",
        delay: int = 0,
    ) -> AnimationTimer:
        """Trigger a fade-out animation."""
        return self["windowOpacity"].transition_to(
            0.0,
            easing=easing,
            duration=duration,
            delay=delay,
        )

    def zoom(
        self,
        duration: int = 1000,
        start: float = 1.0,
        end: float = 1.0,
        easing: core.easingcurve.TypeStr = "in_out_sine",
        anchor: str = "center",
        delay: int = 0,
    ) -> AnimationTimer:
        """Trigger a zoom animation with given anchor."""
        from prettyqt import animations

        anim = animations.ZoomAnimation(parent=self._widget, anchor=anchor)
        anim.set_easing(easing)
        anim.set_start_value(start)
        anim.set_end_value(end)
        anim.set_duration(duration)
        self.run(anim, delay)
        return anim

    def slide(
        self,
        duration: int = 1000,
        start=None,
        end=None,
        easing: core.easingcurve.TypeStr = "in_out_sine",
        delay: int = 0,
        reverse: bool = False,
        single_shot: bool = True,
    ) -> AnimationTimer:
        anim = core.PropertyAnimation(parent=self._widget)
        anim.set_easing(easing)
        pos = self._widget.geometry().topLeft()
        start_offset = core.Point(0, 0) if start is None else datatypes.to_point(start)
        end_offset = core.Point(0, 0) if end is None else datatypes.to_point(end)
        anim.set_start_value(pos + start_offset)
        anim.set_end_value(pos + end_offset)
        anim.setDuration(duration)
        anim.apply_to(self._widget.pos)
        if reverse:
            anim.append_reversed()
        return self.run(anim, delay, single_shot=single_shot)

    def bounce(
        self,
        end: datatypes.PointType,
        easing: core.easingcurve.TypeStr = "in_out_sine",
        duration: int = 1000,
        delay: int = 0,
    ) -> AnimationTimer:
        """Trigger a move animation to given offset and return to original position."""
        return self["pos"].transition_to(
            end,
            easing=easing,
            duration=duration,
            delay=delay,
            reverse=True,
        )

    def run(
        self,
        animation: core.QPropertyAnimation,
        delay: int | str = 0,
        single_shot: bool = True,
    ) -> AnimationTimer:
        """Run an animation with given delay.

        Arguments:
            animation: Animation to run
            delay: delay after which animation should start
            single_shot: whethere the animation should trigger once or repeat

        Returns:
            AnimationTimer, a core.Timer subclass with the PropertyAnimation attached.
        """
        if not animation.targetObject().isVisible():
            logger.info("Attention. Starting animation for invisible widget.")
        logger.debug(f"starting {animation!r} with {delay=}. ({single_shot=})")
        timer = AnimationTimer(
            parent=self._widget,
            single_shot=single_shot,
            interval=helpers.parse_time(delay) if isinstance(delay, str) else delay,
            animation=animation,
        )
        timer.start()
        return timer

    def highlight_widget(self, widget: widgets.QWidget):
        from prettyqt.custom_widgets.overlayborder import FocusWidget

        widget = FocusWidget(self, widget)
        widget.show()


if __name__ == "__main__":
    app = widgets.app()
    with app.debug_mode():
        w1 = widgets.RadioButton("test")
        w2 = widgets.RadioButton("test")
        w3 = widgets.Splitter("horizontal")
        w4 = widgets.RadioButton("test")
        w5 = widgets.PlainTextEdit("test")
        widget = widgets.Widget()
        container = widget.set_layout("horizontal")
        container.add(w1)
        container.add(w2)
        w3.add(w4)
        w3.add(w5)
        container.add(w3)
        widget.show()
        # container[:].fx["size"].transition_from(
        #     (100, 100), duration=3000, reverse=True, single_shot=False
        # )
        # timer = container[::2].fx.slide(
        #     start=(-100, 0), duration=3000, reverse=True, single_shot=False
        # )
        container[:].fx["pos"].animate_on_event(
            "hover_enter", end=(0, -100), duration=1000, start=None
        )
        w5.fx.bounce((0, -100), duration=2000, delay=5000)
        # print(timer)
        # app.sleep(1)  #
        # timer[0].animation.toggle_direction()
        app.exec()



from __future__ import annotations

from prettyqt.qt.QtPdfWidgets import *  # noqa: F403

from .pdfview import PdfView
from prettyqt.qt import QtPdfWidgets

QT_MODULE = QtPdfWidgets

__all__ = ["PdfView"]



"""Provides a widget class for displaying a Qt Quick user interface."""

from __future__ import annotations

from prettyqt.qt.QtQuickWidgets import *  # noqa: F403

from .quickwidget import QuickWidget
from prettyqt.qt import QtQuickWidgets

QT_MODULE = QtQuickWidgets

__all__ = [
    "QuickWidget",
]



from __future__ import annotations

from typing import Literal

from prettyqt import core
from prettyqt.qt import QtSpatialAudio
from prettyqt.utils import bidict


OutputModeStr = Literal["infinite", "once"]

OUTPUT_MODE: bidict[OutputModeStr, QtSpatialAudio.QAudioEngine.OutputMode] = bidict(
    surround=QtSpatialAudio.QAudioEngine.OutputMode.Surround,
    stereo=QtSpatialAudio.QAudioEngine.OutputMode.Stereo,
    headphone=QtSpatialAudio.QAudioEngine.OutputMode.Headphone,
)


class AudioEngine(core.ObjectMixin, QtSpatialAudio.QAudioEngine):
    """Manages a three dimensional sound field."""

    def set_output_mode(
        self, mode: OutputModeStr | QtSpatialAudio.QAudioEngine.OutputMode
    ):
        """Set the output mode.

        Args:
            mode: output mode
        """
        self.setOutputMode(OUTPUT_MODE.get_enum_value(mode))

    def get_output_mode(self) -> OutputModeStr:
        """Return current output mode.

        Returns:
            output mode
        """
        return OUTPUT_MODE.inverse[self.outputMode()]


if __name__ == "__main__":
    engine = AudioEngine()
    engine.set_output_mode("infinite")



from __future__ import annotations

from collections.abc import Sequence
import dataclasses
import logging
from typing import Any

from prettyqt import constants, core, itemmodels
from prettyqt.utils import datatypes


logger = logging.getLogger(__name__)


class DataClassModel(itemmodels.BaseDataclassModel):
    """Table model to display a list of dataclasses."""

    SUPPORTS = Sequence[datatypes.IsDataclass]

    def __init__(self, items: Sequence[datatypes.IsDataclass], **kwargs):
        super().__init__(items, **kwargs)

    @classmethod
    def supports(cls, instance) -> bool:
        match instance:
            case (datatypes.IsDataclass(), *_):
                return True
            case _:
                return False

    def get_fields(self) -> dict[str, Any]:
        fields = dataclasses.fields(self.Class)
        return {field.name: field for field in fields}

    def flags(self, parent: core.ModelIndex) -> constants.ItemFlag:
        if not parent.isValid():
            return super().flags(parent)
        if self.Class.__dataclass_params__.frozen:
            super().flags(parent)
        field_name = self._field_names[parent.column()]
        instance = self.items[parent.row()]
        # need to cover not parent.isValid()?
        val = getattr(instance, field_name)
        if isinstance(val, bool):
            return super().flags(parent) | constants.IS_CHECKABLE
        return super().flags(parent) | constants.IS_EDITABLE


if __name__ == "__main__":
    from prettyqt import widgets

    app = widgets.app()

    @dataclasses.dataclass
    class BaseSetting:
        name: str
        description: str
        some_bool: bool = False
        minimum: float = 0.0
        maximum: float = 0.0

    one = BaseSetting(
        name="Setting 1",
        description="Description of setting 1",
        some_bool=True,
        minimum=5.0,
    )

    two = BaseSetting(
        name="Setting 2",
        description="Description of setting 2",
        some_bool=True,
        minimum=50.0,
    )

    items = [one, two]
    klasses = [type(i) for i in items]
    view = widgets.TableView()
    view.set_icon("mdi.folder")
    model = DataClassModel(items, parent=view)
    view.set_model(model)
    view.setEditTriggers(view.EditTrigger.AllEditTriggers)
    view.set_delegate("editor")
    view.resize(1000, 1000)
    with app.debug_mode():
        view.show()
        app.exec()



from __future__ import annotations

from prettyqt import gui, widgets


class DefaultDropDownAction(widgets.WidgetAction):
    """This action provides a button as well as a dropdown to choose a default action.

    The default action can be selected from a drop down list. The last one used
    becomes the default one.

    The default action is directly usable without using the drop down list.
    """

    def __init__(self, parent=None, **kwargs):
        super().__init__(parent, **kwargs)
        button = widgets.ToolButton(parent)
        button.set_popup_mode("menu_button")
        self.setDefaultWidget(button)
        self._button = button

    def get_menu(self) -> widgets.Menu:
        """Returns the menu."""
        menu = self._button.menu()
        if menu is None:
            menu = widgets.Menu(self._button)
            self._button.setMenu(menu)
        return menu

    def addAction(self, action: gui.QAction):
        """Add a new action to the list."""
        menu = self.get_menu()
        menu.addAction(action)
        if self._button.defaultAction() is None:
            self._button.setDefaultAction(action)
        if action.isCheckable():
            action.toggled.connect(self._toggled)

    def _toggled(self, checked: bool):
        if checked:
            action = self.sender()
            self._button.setDefaultAction(action)


if __name__ == "__main__":
    app = widgets.app()
    window = widgets.MainWindow()
    toolbar = widgets.ToolBar()
    action = gui.Action("test")
    window.add_toolbar(toolbar, "top")
    widgetaction = DefaultDropDownAction()
    widgetaction.addAction(action)
    toolbar.addAction(widgetaction)
    window.show()
    app.exec()



from __future__ import annotations

from typing import Literal

from prettyqt import widgets
from prettyqt.utils import bidict


ToolBarFeatureStr = Literal["none", "movable"]

TOOLBAR_FEATURE: bidict[ToolBarFeatureStr, widgets.QStyleOptionToolBar.ToolBarFeature] = (
    bidict(
        none=widgets.QStyleOptionToolBar.ToolBarFeature(0),  # type: ignore
        movable=widgets.QStyleOptionToolBar.ToolBarFeature.Movable,
    )
)


class StyleOptionToolBar(widgets.StyleOptionMixin, widgets.QStyleOptionToolBar):
    pass



from __future__ import annotations

import logging
from typing import TYPE_CHECKING, Any

from prettyqt import core


if TYPE_CHECKING:
    from collections.abc import Callable


logger = logging.getLogger(__name__)


class PredicateFilterProxyModel(core.SortFilterProxyModel):
    """A simple filter proxy model with settable filter predicates.

    ### Example:
    ```py
    proxy = PredicateFilterProxyModel()
    proxy.add_filter(lambda value: value < 1)
    ```
    """

    ID = "predicate_filter"

    def __init__(self, parent=None, **kwargs):
        super().__init__(parent, **kwargs)
        self._filters: list[Callable[[Any], bool]] = []

    def clear_filters(self):
        self._filters = []
        self.invalidateFilter()

    def add_filter(self, predicate: Callable[[Any], bool]):
        self._filters.append(predicate)
        self.invalidateFilter()

    def filterAcceptsRow(self, row, parent):
        source = self.sourceModel()
        col = self.filterKeyColumn()
        role = self.filterRole()
        index = source.index(row, col, parent)
        data = source.data(index, role)

        def apply(f: Callable):
            try:
                return f(data)
            except (TypeError, ValueError):
                return False

        return all(apply(f) for f in self._filters)


if __name__ == "__main__":
    from prettyqt import widgets
    from prettyqt.itemmodels import JsonModel

    app = widgets.app()
    dist = [
        dict(
            assss=2,
            bffff={
                "a": 4,
                "b": [1, 2, 3],
                "jkjkjk": "tekjk",
                "sggg": "tekjk",
                "fdfdf": "tekjk",
                "xxxx": "xxx",
            },
        ),
        6,
        "jkjk",
    ]

    _source_model = JsonModel(dist)
    model = PredicateFilterProxyModel()
    model.setFilterKeyColumn(1)
    model.setSourceModel(_source_model)
    table = widgets.TreeView()
    table.setRootIsDecorated(True)
    table.set_model(model)
    table.show()
    with app.debug_mode():
        app.exec()



from __future__ import annotations

import collections
import logging
import pathlib
from typing import TYPE_CHECKING

from prettyqt import constants, core, gui, itemmodels, widgets
from prettyqt.utils import colors, datatypes


if TYPE_CHECKING:
    from collections.abc import Sequence


logger = logging.getLogger(__name__)

COMMANDS = collections.defaultdict(list)


class CommandGroup:
    def __init__(self, group_name: str):
        self.group_name = group_name

    def register(
        self, text: str, icon: datatypes.IconType | None = None, shortcut: str = ""
    ):
        def decorator(fn):
            COMMANDS[self.group_name].append((text, icon, shortcut, fn))
            return fn

        return decorator


class CommandTable(widgets.TableView):
    """TableView showing a Command Palette."""

    action_clicked = core.Signal(int)

    def __init__(self, parent: widgets.Widget | None = None) -> None:
        super().__init__(
            parent,
            selection_mode="single",
            selection_behavior="rows",
            sorting_enabled=True,
        )
        # self.set_cursor("pointing_hand")
        self._model = itemmodels.ActionsModel([], parent=self)
        self._proxy = itemmodels.FuzzyFilterProxyModel(
            filter_key_column=0, invalidated=self.select_first_row
        )
        self._proxy.set_filter_case_sensitive(False)
        # self._proxy.set_sort_role(constants.SORT_ROLE)
        self._proxy.setSourceModel(self._model)
        self.setModel(self._proxy)
        self.h_header.set_resize_mode("stretch")
        self.pressed.connect(self._on_clicked)
        self.set_delegate("html", column=0)
        self._match_color = gui.QColor("#468cc6")
        self.setShowGrid(False)

    @core.Property(gui.QColor)
    def matchColor(self) -> gui.QColor:
        return self._match_color

    @matchColor.setter
    def matchColor(self, color: gui.QColor):
        self._match_color = color

    def _on_clicked(self, index: core.ModelIndex) -> None:
        if index.isValid():
            role = self._model.ExtraRoles.TreeItemRole
            data = index.data(role)
            data.trigger()

    def execute_focused(self):
        fn = self.current_data()
        fn.trigger()


class CommandPalette(widgets.Widget):
    """A Qt command palette widget."""

    def __init__(self, parent: widgets.QWidget | None = None):
        super().__init__(parent=parent)
        self.setWindowFlags(
            constants.WindowType.WindowStaysOnTopHint
            | constants.WindowType.FramelessWindowHint
            # | constants.WindowType.ToolTip
        )
        self.set_focus_policy("strong")
        self.setMinimumWidth(700)
        self._line = widgets.LineEdit()
        self.setFocusProxy(self._line)
        self._table = CommandTable()
        # self._line.value_changed.connect(self._table.select_first_row)
        self._line.value_changed.connect(self._table._proxy.set_search_term)
        layout = self.set_layout("vertical")
        layout.addWidget(self._line)
        layout.addWidget(self._table)
        self.add_shortcut("Ctrl+P", callback=self.close)
        self._line.installEventFilter(self)

        # self._line.textChanged.connect(self._on_text_changed)
        # self._table.action_clicked.connect(self._on_action_clicked)
        # self._line.editingFinished.connect(self.hide)

    def eventFilter(self, source: core.QObject, e: core.QEvent) -> bool:
        if source != self._line or e.type() != core.QEvent.Type.KeyPress:
            return super().eventFilter(source, e)
        if e.modifiers() in (
            constants.KeyboardModifier.NoModifier,
            constants.KeyboardModifier.KeypadModifier,
        ):
            match e.key():
                case constants.Key.Key_Escape:
                    self.hide()
                    return True
                case constants.Key.Key_Return:
                    self.hide()
                    self._table.execute_focused()
                    return True
                case constants.Key.Key_Up:
                    self._table.move_row_selection(-1)
                    return True
                case constants.Key.Key_Down:
                    self._table.move_row_selection(1)
                    return True
        return super().eventFilter(source, e)

    def populate_from_widget(self, widget: widgets.QWidget):
        self.add_actions(widget.actions())
        if not callable(widget.parent):
            return
        while widget := widget.parent():
            self.add_actions(widget.actions())

    def add_path_actions(self, path):
        path = pathlib.Path(path)
        actions = [gui.Action(str(p)) for p in path.rglob("*") if p.is_file()]
        self.add_actions(actions)

    def match_color(self) -> str:
        """The color used for the matched characters."""
        return self._table.match_color

    def set_match_color(self, color):
        """Set the color used for the matched characters."""
        self._table.match_color = colors.get_color(color).name()

    def install_to(self, parent: widgets.Widget):
        self.setParent(parent, constants.WindowType.SubWindow)
        self.hide()

    # def focusOutEvent(self, a0: gui.QFocusEvent) -> None:
    #     self.hide()
    #     return super().focusOutEvent(a0)

    def add_actions(self, actions: Sequence[gui.QAction]):
        self._table._model.add_items(actions)

    def show(self):
        self._line.setText("")
        self.resize(1000, 300)
        self.position_on("screen")
        super().show()
        self.raise_()
        self._line.setFocus()


if __name__ == "__main__":
    import random
    import string

    app = widgets.app()
    window = widgets.MainWindow()
    window.setCentralWidget(widgets.Label("Press Ctrl+P"))
    window.add_action(text="MainWindowAction", parent=window, triggered=window.close)
    pal = CommandPalette()
    window.add_shortcut("Ctrl+P", pal.show)
    actions = [
        gui.Action(
            text="super duper action",
            shortcut="Ctrl+A",
            tool_tip="some Tooltip text",
            icon="mdi.folder",
            triggered=lambda: print("test"),
        ),
        gui.Action(
            text="this is an action",
            shortcut="Ctrl+B",
            tool_tip="Tooltip",
            icon="mdi.folder-outline",
            checked=True,
            checkable=True,
        ),
        gui.Action(
            text="another one P",
            shortcut="Ctrl+Alt+A",
            tool_tip="Some longer tooltiPpp",
            icon="mdi.folder",
        ),
        gui.Action(
            text="another onpe P",
            shortcut="Ctrl+Alt+A",
            tool_tip="Some longer tooltiPpp",
            icon="mdi.folder",
        ),
        gui.Action(text="a", shortcut="Ctrl+A", tool_tip="Tooltip", icon="mdi.folder"),
    ]
    pal.populate_from_widget(window)
    pal.add_actions(actions)
    for _ in range(1000):
        label = "".join(random.choices(string.ascii_uppercase, k=10))
        pal.add_actions([gui.Action(text=label)])
    window.show()
    with app.debug_mode():
        app.exec()



from __future__ import annotations

import sys
from typing import Any, Literal, Self

from prettyqt import gui
from prettyqt.utils import bidict, get_repr


CAPITALIZATION = bidict(
    mixed_case=gui.QFont.Capitalization.MixedCase,
    all_uppercase=gui.QFont.Capitalization.AllUppercase,
    all_lowercase=gui.QFont.Capitalization.AllLowercase,
    small_caps=gui.QFont.Capitalization.SmallCaps,
    capitalize=gui.QFont.Capitalization.Capitalize,
)

CapitalizationStr = Literal[
    "mixed_case", "all_uppercase", "all_lowercase", "small_caps", "capitalize"
]

HINTING_PREFERENCE = bidict(
    default=gui.QFont.HintingPreference.PreferDefaultHinting,
    none=gui.QFont.HintingPreference.PreferNoHinting,
    vertical=gui.QFont.HintingPreference.PreferVerticalHinting,
    full=gui.QFont.HintingPreference.PreferFullHinting,
)

HintingPreferenceStr = Literal["default", "none", "vertical", "full"]

SPACING_TYPE = bidict(
    percentage=gui.QFont.SpacingType.PercentageSpacing,
    absolute=gui.QFont.SpacingType.AbsoluteSpacing,
)

SpacingTypeStr = Literal["percentage", "absolute"]

STRETCH = bidict(
    any=gui.QFont.Stretch.AnyStretch,
    ultra_condensed=gui.QFont.Stretch.UltraCondensed,
    extra_condensed=gui.QFont.Stretch.ExtraCondensed,
    condensed=gui.QFont.Stretch.Condensed,
    semi_condensed=gui.QFont.Stretch.SemiCondensed,
    unstretched=gui.QFont.Stretch.Unstretched,
    semi_expanded=gui.QFont.Stretch.SemiExpanded,
    expanded=gui.QFont.Stretch.Expanded,
    extra_expanded=gui.QFont.Stretch.ExtraExpanded,
    ultra_expanded=gui.QFont.Stretch.UltraExpanded,
)

StretchStr = Literal[
    "any",
    "ultra_condensed",
    "extra_condensed",
    "condensed",
    "semi_condensed",
    "unstretched",
    "semi_expanded",
    "expanded",
    "extra_expanded",
    "ultra_expanded",
]

STYLE = bidict(
    normal=gui.QFont.Style.StyleNormal,
    italic=gui.QFont.Style.StyleItalic,
    oblique=gui.QFont.Style.StyleOblique,
)

StyleStr = Literal["normal", "italic", "oblique"]

STYLE_STRATEGY = bidict(
    prefer_default=gui.QFont.StyleStrategy.PreferDefault,
    prefer_bitmap=gui.QFont.StyleStrategy.PreferBitmap,
    prefer_device=gui.QFont.StyleStrategy.PreferDevice,
    prefer_outline=gui.QFont.StyleStrategy.PreferOutline,
    force_outline=gui.QFont.StyleStrategy.ForceOutline,
    no_antialias=gui.QFont.StyleStrategy.NoAntialias,
    so_subpixel_antialias=gui.QFont.StyleStrategy.NoSubpixelAntialias,
    prefer_antialias=gui.QFont.StyleStrategy.PreferAntialias,
    no_font_merging=gui.QFont.StyleStrategy.NoFontMerging,
    prefer_no_shaping=gui.QFont.StyleStrategy.PreferNoShaping,
)  # ORed with PreferMatch, PreferQuality, ForceIntegerMetrics

StyleStrategyStr = Literal[
    "prefer_default",
    "prefer_bitmap",
    "prefer_device",
    "prefer_outline",
    "force_outline",
    "no_antialias",
    "so_subpixel_antialias",
    "prefer_antialias",
    "no_font_merging",
    "prefer_no_shaping",
]

STYLE_HINTS = bidict(
    any=gui.QFont.StyleHint.AnyStyle,
    sans_serif=gui.QFont.StyleHint.SansSerif,
    serif=gui.QFont.StyleHint.Serif,
    typewriter=gui.QFont.StyleHint.TypeWriter,
    decorative=gui.QFont.StyleHint.Decorative,
    monospace=gui.QFont.StyleHint.Monospace,
    fantasy=gui.QFont.StyleHint.Fantasy,
    cursive=gui.QFont.StyleHint.Cursive,
    system=gui.QFont.StyleHint.System,
)

StyleHintStr = Literal[
    "any",
    "sans_serif",
    "serif",
    "typewriter",
    "decorative",
    "monospace",
    "fantasy",
    "cursive",
    "system",
]

WEIGHT = bidict(
    thin=gui.QFont.Weight.Thin,
    extra_light=gui.QFont.Weight.ExtraLight,
    light=gui.QFont.Weight.Light,
    normal=gui.QFont.Weight.Normal,
    medium=gui.QFont.Weight.Medium,
    demi_bold=gui.QFont.Weight.DemiBold,
    bold=gui.QFont.Weight.Bold,
    extra_bold=gui.QFont.Weight.ExtraBold,
    black=gui.QFont.Weight.Black,
)

WeightStr = Literal[
    "thin",
    "extra_light",
    "light",
    "normal",
    "medium",
    "demi_bold",
    "bold",
    "extra_bold",
    "black",
]


class Font(gui.QFont):
    """Specifies a query for a font used for drawing text."""

    def __repr__(self):
        return get_repr(
            self, self.family(), self.pointSize(), self.weight(), self.italic()
        )

    def __getstate__(self):
        return dict(
            family=self.family(),
            pointsize=self.pointSize(),
            weight=self.weight(),
            italic=self.italic(),
        )

    def __setstate__(self, state):
        self.setFamily(state["family"])
        if state["pointsize"] > -1:
            self.setPointSize(state["pointsize"])
        self.setWeight(state["weight"])
        self.setItalic(state["italic"])

    @property
    def _family(self) -> str:
        return self.family()

    __match_args__ = ("_family",)

    def __reduce__(self):
        return type(self), (), self.__getstate__()

    def serialize(self) -> dict[str, Any]:
        return self.__getstate__()

    @property
    def metrics(self) -> gui.FontMetrics:
        return gui.FontMetrics(self)

    def set_size(self, size: int):
        self.setPointSize(size)

    @classmethod
    def mono(cls, size=8, as_qt: bool = False) -> Self:
        match sys.platform:
            case "win32":
                font = "Consolas"
            case "darwin":
                font = "Menlo"
            case _:
                font = "Monospace"
        if as_qt:
            return gui.QFont(font)
        return cls(font, size)
        # font.setStyleHint()

    def scaled(self, factor: float) -> Self:
        scaled = type(self)(self)
        if self.pointSizeF() != -1:
            scaled.setPointSizeF(self.pointSizeF() * factor)
        elif self.pixelSize() != -1:
            scaled.setPixelSize(int(self.pixelSize() * factor))
        return scaled

    def set_style_hint(self, hint: StyleHintStr | gui.QFont.StyleHint):
        """Set the style hint.

        Args:
            hint: style hint
        """
        self.setStyleHint(STYLE_HINTS.get_enum_value(hint))

    def set_weight(self, weight: WeightStr | gui.QFont.Weight):
        """Set the font weight.

        Args:
            weight: font weight
        """
        self.setWeight(WEIGHT.get_enum_value(weight))

    def get_weight(self) -> WeightStr:
        """Get current font weight.

        Returns:
            current font weight
        """
        return WEIGHT.inverse[self.weight()]

    def set_capitalization(
        self, capitalization: CapitalizationStr | gui.Font.Capitalization
    ):
        """Set the font capitalization.

        Args:
            capitalization: font capitalization
        """
        self.setCapitalization(CAPITALIZATION.get_enum_value(capitalization))

    def get_capitalization(self) -> CapitalizationStr:
        """Get current font capitalization.

        Returns:
            current font capitalization
        """
        return CAPITALIZATION.inverse[self.capitalization()]

    def set_hinting_preference(
        self, preference: HintingPreferenceStr | gui.Font.HintingPreference
    ):
        """Set the hinting preference.

        Args:
            preference: hinting preference
        """
        self.setHintingPreference(HINTING_PREFERENCE.get_enum_value(preference))

    def get_hinting_preference(self) -> HintingPreferenceStr:
        """Get current hinting preference.

        Returns:
            current hinting preference
        """
        return HINTING_PREFERENCE.inverse[self.hintingPreference()]

    def set_letter_spacing(
        self, typ: SpacingTypeStr | gui.Font.SpacingType, spacing: float
    ):
        """Set the letter spacing.

        Args:
            typ: letter spacing type
            spacing: spacing
        """
        self.setLetterSpacing(SPACING_TYPE.get_enum_value(typ), spacing)

    def get_letter_spacing_type(self) -> SpacingTypeStr:
        """Get current letter spacing type.

        Returns:
            current letter spacing type
        """
        return SPACING_TYPE.inverse[self.letterSpacingType()]

    def set_style(self, style: StyleStr | gui.Font.Style):
        """Set the font style.

        Args:
            style: font style
        """
        self.setStyle(STYLE.get_enum_value(style))

    def get_style(self) -> StyleStr:
        """Get current font style.

        Returns:
            current font style
        """
        return STYLE.inverse[self.style()]

    def set_family(self, family: str, fallback: str | None = None):
        """Set the font family.

        Args:
            family: font family
            fallback: fallback font family
        """
        self.setFamily(family)
        font_info = gui.FontInfo(self)
        if fallback is not None and font_info.family() != family:
            self.setFamily(fallback)


if __name__ == "__main__":
    font = Font("Consolas")
